"""SDK functions for managed jobs."""
import ipaddress
import os
import pathlib
import tempfile
import typing
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib import parse as urlparse
import uuid

import colorama

from sky import backends
from sky import core
from sky import exceptions
from sky import execution
from sky import global_user_state
from sky import optimizer as optimizer_lib
from sky import provision as provision_lib
from sky import sky_logging
from sky import skypilot_config
from sky import task as task_lib
from sky.adaptors import common as adaptors_common
from sky.backends import backend_utils
from sky.backends import cloud_vm_ray_backend
from sky.catalog import common as service_catalog_common
from sky.dag import DEFAULT_EXECUTION
from sky.data import data_utils
from sky.data import storage as storage_lib
from sky.jobs import constants as managed_job_constants
from sky.jobs import state as managed_job_state
from sky.jobs import utils as managed_job_utils
from sky.metrics import utils as metrics_lib
from sky.provision import common as provision_common
from sky.schemas.api import responses
from sky.serve import serve_state
from sky.serve import serve_utils
from sky.serve.server import impl
from sky.server.requests import request_names
from sky.skylet import constants as skylet_constants
from sky.usage import usage_lib
from sky.utils import admin_policy_utils
from sky.utils import common
from sky.utils import common_utils
from sky.utils import controller_utils
from sky.utils import dag_utils
from sky.utils import rich_utils
from sky.utils import status_lib
from sky.utils import subprocess_utils
from sky.utils import timeline
from sky.utils import ux_utils
from sky.workspaces import core as workspaces_core

if typing.TYPE_CHECKING:
    from google.protobuf import json_format

    import sky
    from sky import resources
    from sky.schemas.generated import managed_jobsv1_pb2
else:
    json_format = adaptors_common.LazyImport('google.protobuf.json_format')

    managed_jobsv1_pb2 = adaptors_common.LazyImport(
        'sky.schemas.generated.managed_jobsv1_pb2')

logger = sky_logging.init_logger(__name__)

_MANAGED_JOB_FIELDS_FOR_QUEUE_KUBERNETES = [
    'job_id',
    'task_id',
    'workspace',
    'job_name',
    'task_name',
    'resources',
    'submitted_at',
    'end_at',
    'job_duration',
    'recovery_count',
    'status',
    'pool',
    'current_cluster_name',
    'job_id_on_pool_cluster',
    'start_at',
    'infra',
    'cloud',
    'region',
    'zone',
    'cluster_resources',
    'schedule_state',
    'details',
    'failure_reason',
    'metadata',
    'user_name',
    'user_hash',
]


def _warn_file_mounts_rolling_update(dag: 'sky.Dag') -> None:
    """Warn if local file mounts or workdir may be lost during rolling update.

    When rolling update is enabled with consolidation mode but no jobs bucket
    is configured, local file mounts and workdirs are stored locally on the API
    server pod and will be lost during a rolling update.
    """
    # If rolling update is not enabled, don't warn.
    if os.environ.get(skylet_constants.SKYPILOT_ROLLING_UPDATE_ENABLED) is None:
        return

    # If persistent storage is enabled (via Helm storage.enabled=true or by
    # default for local deployments), file mounts are persisted and will
    # survive rolling updates. Default to True if not explicitly set to False.
    storage_enabled_str = os.environ.get(
        skylet_constants.SKYPILOT_API_SERVER_STORAGE_ENABLED, 'true')
    if storage_enabled_str.lower() == 'true':
        return

    # If consolidation mode is not enabled, don't warn.
    if not managed_job_utils.is_consolidation_mode():
        return

    # If a jobs bucket is configured, don't warn.
    if skypilot_config.get_nested(('jobs', 'bucket'), None) is not None:
        return

    # Check if any task has local file_mounts (not cloud store URLs) or workdir
    has_local_file_mounts = False
    has_local_workdir = False
    for task_ in dag.tasks:
        if task_.file_mounts:
            for src in task_.file_mounts.values():
                if not data_utils.is_cloud_store_url(src):
                    has_local_file_mounts = True
                    break
        if task_.workdir and isinstance(task_.workdir, str):
            has_local_workdir = True
            break
        if has_local_file_mounts:
            break

    if not has_local_file_mounts and not has_local_workdir:
        return

    logger.warning(
        f'{colorama.Fore.YELLOW}WARNING: Local file mounts or workdir detected '
        'with rolling update enabled for API server. To persist files'
        ' across API server restarts/update, use buckets, volumes, or git '
        'for your file mounts; or, configure a bucket in your SkyPilot config '
        'under `jobs.bucket`; or, enable persistent storage in Helm with '
        '`storage.enabled=true`. See: https://docs.skypilot.co/en/latest/'
        'reference/kubernetes/kubernetes-deployment.html'
        f'{colorama.Style.RESET_ALL}')


def _upload_files_to_controller(dag: 'sky.Dag') -> Dict[str, str]:
    """Upload files to the controller.

    In consolidation mode, we still need to upload files to the controller as
    we should keep a separate workdir for each jobs. Assuming two jobs using
    the same workdir, if there are some modifications to the workdir after job 1
    is submitted, on recovery of job 1, the modifications should not be applied.
    """
    local_to_controller_file_mounts: Dict[str, str] = {}

    # Check if user has explicitly configured a bucket for jobs.
    # If so, we should use cloud storage even in consolidation mode to persist
    # files across rolling updates and pod restarts.
    has_explicit_bucket = skypilot_config.get_nested(('jobs', 'bucket'),
                                                     None) is not None
    storage_clouds = (
        storage_lib.get_cached_enabled_storage_cloud_names_or_refresh())
    force_disable_cloud_bucket = skypilot_config.get_nested(
        ('jobs', 'force_disable_cloud_bucket'), False)
    # Use cloud storage if:
    # 1. Not in consolidation mode, OR
    # 2. In consolidation mode BUT user has explicit bucket configured
    # AND storage clouds are available AND cloud bucket is not force-disabled
    if ((not managed_job_utils.is_consolidation_mode() or has_explicit_bucket)
            and storage_clouds and not force_disable_cloud_bucket):
        for task_ in dag.tasks:
            controller_utils.maybe_translate_local_file_mounts_and_sync_up(
                task_, task_type='jobs')
    else:
        # We do not have any cloud storage available, so fall back to
        # two-hop file_mount uploading.
        # Note: we can't easily hack sync_storage_mounts() to upload
        # directly to the controller, because the controller may not
        # even be up yet.
        for task_ in dag.tasks:
            if task_.storage_mounts and not storage_clouds:
                # Technically, we could convert COPY storage_mounts that
                # have a local source and do not specify `store`, but we
                # will not do that for now. Only plain file_mounts are
                # supported.
                raise exceptions.NotSupportedError(
                    'Cloud-based file_mounts are specified, but no cloud '
                    'storage is available. Please specify local '
                    'file_mounts only.')

            # Merge file mounts from all tasks.
            local_to_controller_file_mounts.update(
                controller_utils.translate_local_file_mounts_to_two_hop(task_))

    return local_to_controller_file_mounts


def _job_ids_to_str(job_ids: Optional[List[int]]) -> str:
    if not job_ids:
        return ''

    if len(job_ids) == 1:
        return str(job_ids[0])

    job_ids = sorted(job_ids)
    ranges = []
    start = prev = job_ids[0]

    for n in job_ids[1:]:
        if n == prev + 1:
            prev = n
            continue
        ranges.append(f'{start}-{prev}' if start != prev else str(start))
        start = prev = n

    # append last range
    ranges.append(f'{start}-{prev}' if start != prev else str(start))
    return ','.join(ranges)


def _consolidated_launch(
    controller: controller_utils.Controllers,
    controller_task: 'sky.Task',
    job_ids: List[int],
) -> Tuple[List[int], backends.ResourceHandle]:
    local_handle = backend_utils.is_controller_accessible(controller=controller,
                                                          stopped_message='')
    backend = backend_utils.get_backend_from_handle(local_handle)
    assert isinstance(backend, backends.CloudVmRayBackend)
    with sky_logging.silent():
        backend.sync_file_mounts(handle=local_handle,
                                 all_file_mounts=controller_task.file_mounts,
                                 storage_mounts=controller_task.storage_mounts)
    run_script = controller_task.run
    assert isinstance(run_script, str)
    # Manually add the env variables to the run script.
    # Originally this is done in ray jobs submission but now
    # we have to do it manually because there is no ray
    # runtime on the API server.
    env_cmds = [f'export {k}={v!r}' for k, v in controller_task.envs.items()]
    run_script = '\n'.join(env_cmds + [run_script])
    # Dump script for high availability recovery.
    assert job_ids is not None, 'job_ids not set'
    log_dir = os.path.join(skylet_constants.SKY_LOGS_DIRECTORY, 'managed_jobs')
    os.makedirs(log_dir, exist_ok=True)
    job_ids_str = _job_ids_to_str(job_ids)
    log_path = os.path.join(log_dir, f'submit-job-{job_ids_str}.log')
    backend.run_on_head(local_handle, run_script, log_path=log_path)
    ux_utils.starting_message(f'Job submitted, ID: {job_ids_str}')
    return job_ids, local_handle


def _maybe_submit_job_locally(prefix: str, dag: 'sky.Dag',
                              num_jobs: int) -> Optional[List[int]]:
    """Submit the managed job locally if in consolidation mode.

    In normal mode the managed job submission is done in the ray job submission.
    For consolidation mode, we need to manually submit it. Check the following
    function for the normal mode submission:
    sky/backends/cloud_vm_ray_backend.py::CloudVmRayBackend,
    _exec_code_on_head::_maybe_add_managed_job_code
    """
    if not managed_job_utils.is_consolidation_mode():
        return None

    # Create local directory for the managed job.
    pathlib.Path(prefix).expanduser().mkdir(parents=True, exist_ok=True)
    job_ids = []
    pool = dag.pool
    pool_hash = None
    if pool is not None:
        pool_hash = serve_state.get_service_hash(pool)
        # Already checked in the sdk.
        assert pool_hash is not None, f'Pool {pool} not found'
    for _ in range(num_jobs):
        # TODO(tian): We should have a separate name for each job when
        # submitting multiple jobs. Current blocker is that we are sharing
        # the same dag object for all jobs. Maybe we can do copy.copy() for
        # each job and then give it a unique name (e.g. append job id after
        # the task name). The name of the dag also needs to be aligned with
        # the task name.
        # Execution mode: 'parallel' for job groups, 'serial' for pipelines and
        # single jobs
        execution_mode = (dag.execution.value
                          if dag.execution else DEFAULT_EXECUTION.value)
        consolidation_mode_job_id = (
            managed_job_state.set_job_info_without_job_id(
                dag.name,
                workspace=skypilot_config.get_active_workspace(
                    force_user_workspace=True),
                entrypoint=common_utils.get_current_command(),
                pool=pool,
                pool_hash=pool_hash,
                user_hash=common_utils.get_user_hash(),
                execution=execution_mode))
        for task_id, task in enumerate(dag.tasks):
            resources_str = backend_utils.get_task_resources_str(
                task, is_managed_job=True)
            # For job groups, determine which tasks are primary vs auxiliary.
            # For non-job-groups (single jobs, pipelines),
            # is_primary_in_job_group is None for all tasks.
            is_primary_in_job_group: Optional[bool] = None
            if dag.is_job_group():
                is_primary_in_job_group = (dag.primary_tasks is None or
                                           task.name in dag.primary_tasks)
            managed_job_state.set_pending(consolidation_mode_job_id, task_id,
                                          task.name, resources_str,
                                          task.metadata_json,
                                          is_primary_in_job_group)
        job_ids.append(consolidation_mode_job_id)
    return job_ids


def _ensure_controller_up(
    controller: controller_utils.Controllers,
    task_resources: Optional[List['resources.Resources']] = None
) -> 'cloud_vm_ray_backend.CloudVmRayResourceHandle':
    """Ensure the jobs controller is up before proceeding.

    If the controller is not accessible, provision it (bring up the cluster)
    without launching a job. This avoids creating a cluster job ID that would
    interfere with the ID space from the controller's perspective.

    Args:
        controller: The controller type to ensure is up.
        task_resources: Optional list of task resources. If provided, the
            controller will be launched on the same cloud as the tasks.
    """
    controller_name = controller.value.cluster_name
    logger.info(f'{colorama.Fore.YELLOW}'
                f'Ensuring the jobs controller {controller_name} is up before'
                f' continuing job launch...{colorama.Style.RESET_ALL}')

    # Create a minimal task for provisioning the controller cluster
    # We only use this for its resources, not to execute a job
    # Use task_resources to determine which cloud to launch the controller.
    controller_resources_set = controller_utils.get_controller_resources(
        controller=controller,
        task_resources=task_resources if task_resources else [])

    # Use the jobs controller template to ensure cloud dependencies
    # are installed.
    dag_uuid = str(uuid.uuid4())

    vars_to_fill: Dict[str, Any] = {
        'dag_name': 'ensure_controller_up',
        'job_controller_indicator_file':
            managed_job_constants.JOB_CONTROLLER_INDICATOR_FILE,
        **controller_utils.controller_only_vars_to_fill(controller,),
    }

    yaml_path = os.path.join(managed_job_constants.JOBS_CONTROLLER_YAML_PREFIX,
                             f'ensure-controller-up-{dag_uuid}.yaml')

    # Fill the template to create the controller task YAML for provisioning.
    common_utils.fill_template(
        managed_job_constants.JOBS_CONTROLLER_PROVISION_TEMPLATE,
        vars_to_fill,
        output_path=yaml_path)

    # Create task from the template-generated YAML.
    controller_task = task_lib.Task.from_yaml(yaml_path)
    controller_task.set_resources(controller_resources_set)

    with skypilot_config.local_active_workspace_ctx(
            skylet_constants.SKYPILOT_DEFAULT_WORKSPACE):
        with common.with_server_user():
            # Only provision the controller, don't execute a job.
            # Job controller is not placed in kueue, as the
            # controller pod is considered a "system" pod
            # and is not subject to queue limits or preemption.
            with skypilot_config.remove_queue_name_from_config():
                _, _ = execution.launch(
                    task=controller_task,
                    cluster_name=controller_name,
                    retry_until_up=True,
                    stream_logs=False,
                    _request_name=request_names.AdminPolicyRequestName.
                    JOBS_LAUNCH_CONTROLLER,
                    _disable_controller_check=True,
                    fast=True)

    # Verify the controller is now accessible
    handle = backend_utils.is_controller_accessible(controller=controller,
                                                    stopped_message='')
    return handle


def _submit_remotely(controller: controller_utils.Controllers,
                     dag: 'sky.Dag',
                     pool: Optional[str] = None,
                     num_jobs: int = 1) -> List[int]:
    # Ensure the controller is up before trying to create job IDs
    # Use the same cloud as the tasks for the controller
    task_resources = None
    for task in dag.tasks:
        if task.resources:
            task_resources = list(task.resources)
            break
    local_handle = _ensure_controller_up(controller,
                                         task_resources=task_resources)
    backend = backend_utils.get_backend_from_handle(local_handle)
    assert isinstance(backend, backends.CloudVmRayBackend)

    workspace = skypilot_config.get_active_workspace(force_user_workspace=True)
    entrypoint = common_utils.get_current_command()
    pool_hash = serve_state.get_service_hash(pool)
    user_hash = common_utils.get_user_hash()

    # Prepare task data
    task_ids = []
    task_names = []
    metadata_jsons = []
    is_primary_in_job_groups = []
    for task_id, task in enumerate(dag.tasks):
        task_ids.append(task_id)
        assert task.name is not None, 'task name is not set'
        task_names.append(task.name)
        assert task.metadata_json is not None, 'task metadata is not set'
        metadata_jsons.append(task.metadata_json)
        if dag.is_job_group():
            is_primary_in_job_group = (dag.primary_tasks is None or
                                       task.name in dag.primary_tasks)
            is_primary_in_job_groups.append(is_primary_in_job_group)
        else:
            is_primary_in_job_groups.append(False)

    # Use the same resources_str for all tasks
    resources_str = backend_utils.get_task_resources_str(dag.tasks[0],
                                                         is_managed_job=True)

    assert dag.name is not None, 'dag name is not set'
    execution_mode = (dag.execution.value
                      if dag.execution else DEFAULT_EXECUTION.value)
    job_ids = backend.set_job_info_without_job_id(
        handle=local_handle,
        name=dag.name,
        workspace=workspace,
        entrypoint=entrypoint,
        pool=pool,
        pool_hash=pool_hash,
        user_hash=user_hash,
        task_ids=task_ids,
        task_names=task_names,
        resources_str=resources_str,
        metadata_jsons=metadata_jsons,
        num_jobs=num_jobs,
        execution=execution_mode,
        is_primary_in_job_groups=(is_primary_in_job_groups))
    return job_ids


@timeline.event
@usage_lib.entrypoint
def launch(
    task: Union['sky.Task', 'sky.Dag'],
    name: Optional[str] = None,
    pool: Optional[str] = None,
    num_jobs: Optional[int] = None,
    stream_logs: bool = True,
) -> Tuple[Optional[Union[int, List[int]]], Optional[backends.ResourceHandle]]:
    # NOTE(dev): Keep the docstring consistent between the Python API and CLI.
    """Launches a managed job.

    Please refer to sky.cli.job_launch for documentation.

    Args:
        task: sky.Task, or sky.Dag (experimental; 1-task only) to launch as a
          managed job.
        name: Name of the managed job.

    Raises:
        ValueError: cluster does not exist. Or, the entrypoint is not a valid
            chain dag.
        sky.exceptions.NotSupportedError: the feature is not supported.
        sky.exceptions.CachedClusterUnavailable: cached jobs controller cluster
            is unavailable

    Returns:
      job_id: Optional[int]; the job ID of the submitted job. None if the
        backend is not CloudVmRayBackend, or no job is submitted to
        the cluster.
      handle: Optional[backends.ResourceHandle]; handle to the controller VM.
        None if dryrun.
    """
    entrypoint = task
    # using hasattr instead of isinstance to avoid importing sky
    if hasattr(task, 'metadata'):
        metadata = task.metadata
    else:
        # we are a Dag, not a Task
        if len(task.tasks) == 1:
            metadata = task.tasks[0].metadata
        else:
            # doesn't make sense to have a git commit since there might be
            # different metadatas for each task
            metadata = {}

    dag_uuid = str(uuid.uuid4().hex[:4])
    dag = dag_utils.convert_entrypoint_to_dag(entrypoint)

    # Always apply the policy again here, even though it might have been applied
    # in the CLI. This is to ensure that we apply the policy to the final DAG
    # and get the mutated config.
    dag, mutated_user_config = admin_policy_utils.apply(
        dag, request_name=request_names.AdminPolicyRequestName.JOBS_LAUNCH)
    dag.resolve_and_validate_volumes()
    if not dag.is_chain() and not dag.is_job_group():
        with ux_utils.print_exception_no_traceback():
            raise ValueError('Only single-task, chain DAG, or JobGroup is '
                             f'allowed for job_launch. Dag: {dag}')
    if dag.is_job_group() and pool is not None:
        with ux_utils.print_exception_no_traceback():
            raise ValueError('JobGroups do not support pools. Please remove '
                             'the --pool argument when launching a job group.')
    dag.validate()
    # TODO(aylei): use consolidated job controller instead of performing
    # pre-mount operations when submitting jobs.
    dag.pre_mount_volumes()

    # Optimize JobGroup before sending to controller
    # This pre-determines cloud+region for all tasks, enabling parallel launch
    if dag.is_job_group():
        dag = optimizer_lib.Optimizer.optimize_job_group(dag)
        # Apply optimized cloud/region to task resources so they persist
        # through serialization. Without this, each task would be re-optimized
        # independently on the controller, potentially ending up on different
        # infrastructure.
        # TODO(zhwu): make the optimizer aware of multiple jobs directly during
        # the re-optimization, instead of independently.
        for task_ in dag.tasks:
            if task_.best_resources is not None:
                best_cloud = task_.best_resources.cloud
                best_region = task_.best_resources.region
                if best_cloud is not None or best_region is not None:
                    override_params: Dict[str, Any] = {}
                    if best_cloud is not None:
                        override_params['cloud'] = best_cloud
                    if best_region is not None:
                        override_params['region'] = best_region
                    task_.set_resources_override(override_params)

        # Warn if job group is not running on Kubernetes (networking won't work)
        first_task = dag.tasks[0]
        if first_task.best_resources is not None:
            best_cloud = first_task.best_resources.cloud
            if best_cloud is not None and str(
                    best_cloud).lower() != 'kubernetes':
                logger.warning(
                    f'{colorama.Fore.YELLOW}Job group service discovery '
                    f'(hostname-based networking) is only supported on '
                    f'Kubernetes. Tasks will run on {best_cloud} but cannot '
                    f'communicate with each other using hostnames.'
                    f'{colorama.Style.RESET_ALL}')

    # If there is a local postgres db, when the api server tries launching on
    # the remote jobs controller it will fail. therefore, we should remove this
    # before sending the config to the jobs controller.
    # TODO(luca) there are a lot of potential problems with postgres being sent
    # to the jobs controller. for example if the postgres is whitelisted to
    # only the API server, this will then break. the simple solution to that is
    # telling the user to add the jobs controller to the postgres whitelist.
    if not managed_job_utils.is_consolidation_mode():
        db_path = mutated_user_config.get('db', None)
        if db_path is not None:
            parsed = urlparse.urlparse(db_path)
            if ((parsed.hostname == 'localhost' or
                 ipaddress.ip_address(parsed.hostname).is_loopback)):
                mutated_user_config.pop('db', None)

    user_dag_str_user_specified = dag_utils.dump_dag_to_yaml_str(
        dag, use_user_specified_yaml=True)

    dag_utils.maybe_infer_and_fill_dag_and_task_names(dag)

    task_names = set()
    priority = None
    for task_ in dag.tasks:
        if task_.name in task_names:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'Task name {task_.name!r} is duplicated in the DAG. '
                    'Either change task names to be unique, or specify the DAG '
                    'name only and comment out the task names (so that they '
                    'will be auto-generated) .')
        task_names.add(task_.name)

        # Check for priority in resources
        task_priority = None
        if task_.resources:
            # Convert set to list to access elements by index
            resources_list = list(task_.resources)
            # Take first resource's priority as reference
            task_priority = resources_list[0].priority

            # Check all other resources have same priority
            for resource in resources_list[1:]:
                if resource.priority != task_priority:
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError(
                            f'Task {task_.name!r}: All resources must have the '
                            'same priority. Found priority '
                            f'{resource.priority} but expected {task_priority}.'
                        )

        if task_priority is not None:
            if (priority is not None and priority != task_priority):
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        'Multiple tasks in the DAG have different priorities. '
                        'Either specify a priority in only one task, or set '
                        'the same priority for each task.')
            priority = task_priority

    if priority is None:
        priority = skylet_constants.DEFAULT_PRIORITY

    if (priority < skylet_constants.MIN_PRIORITY or
            priority > skylet_constants.MAX_PRIORITY):
        raise ValueError(
            f'Priority must be between {skylet_constants.MIN_PRIORITY}'
            f' and {skylet_constants.MAX_PRIORITY}, got {priority}')

    dag_utils.fill_default_config_in_dag_for_job_launch(dag)

    with rich_utils.safe_status(
            ux_utils.spinner_message('Initializing managed job')):

        # Check whether cached jobs controller cluster is accessible
        cluster_name = (
            controller_utils.Controllers.JOBS_CONTROLLER.value.cluster_name)
        if global_user_state.cluster_with_name_exists(cluster_name):
            # there is a cached jobs controller cluster
            try:
                # TODO: do something with returned status?
                _, _ = backend_utils.refresh_cluster_status_handle(
                    cluster_name=cluster_name,
                    force_refresh_statuses=set(status_lib.ClusterStatus))
            except (exceptions.ClusterOwnerIdentityMismatchError,
                    exceptions.CloudUserIdentityError,
                    exceptions.ClusterStatusFetchingError) as e:
                # we weren't able to refresh the cluster for its status.
                with ux_utils.print_exception_no_traceback():
                    raise exceptions.CachedClusterUnavailable(
                        f'Cached jobs controller cluster '
                        f'{cluster_name} cannot be refreshed. Please check if '
                        'the cluster is accessible. If the cluster was '
                        'removed, consider removing the cluster from SkyPilot '
                        f'with:\n\n`sky down {cluster_name} --purge`\n\n'
                        f'Reason: {common_utils.format_exception(e)}')

    # Warn if file mounts may be lost during rolling update
    _warn_file_mounts_rolling_update(dag)

    local_to_controller_file_mounts = _upload_files_to_controller(dag)
    controller = controller_utils.Controllers.JOBS_CONTROLLER
    controller_name = controller.value.cluster_name
    prefix = managed_job_constants.JOBS_TASK_YAML_PREFIX
    controller_resources = controller_utils.get_controller_resources(
        controller=controller,
        task_resources=sum([list(t.resources) for t in dag.tasks], []))

    if num_jobs and pool is None:
        raise ValueError('Cannot specify num_jobs without pool.')

    num_jobs = num_jobs if num_jobs is not None else 1
    # We do this assignment after applying the admin policy, so that we don't
    # need to serialize the pool name in the dag. The dag object will be
    # preserved. See sky/admin_policy.py::MutatedUserRequest::decode.
    dag.pool = pool
    job_ids = _maybe_submit_job_locally(prefix, dag, num_jobs)
    is_consolidation_mode = job_ids is not None
    if not is_consolidation_mode:
        job_ids = _submit_remotely(controller, dag, pool, num_jobs)
    assert job_ids is not None, 'job_ids is not set'

    # This is only needed for non-consolidation mode. For consolidation
    # mode, the controller uses the same catalog as API server.
    modified_catalogs = {} if is_consolidation_mode else (
        service_catalog_common.get_modified_catalog_file_mounts())

    # Submit the job(s).
    # Create a single set of YAML files (not per-rank)
    remote_orig_user_yaml_path = (
        f'{prefix}/{dag.name}-{dag_uuid}.original_user_yaml')
    remote_user_yaml_path = (f'{prefix}/{dag.name}-{dag_uuid}.yaml')
    remote_user_config_path = (f'{prefix}/{dag.name}-{dag_uuid}.config_yaml')
    remote_env_file_path = (f'{prefix}/{dag.name}-{dag_uuid}.env')

    with tempfile.NamedTemporaryFile(
            prefix=f'managed-dag-{dag.name}-',
            mode='w',
    ) as f, tempfile.NamedTemporaryFile(
            prefix=f'managed-user-dag-{dag.name}-',
            mode='w',
    ) as original_user_yaml_path:
        original_user_yaml_path.write(user_dag_str_user_specified)
        original_user_yaml_path.flush()

        # Set the num_jobs env variable for each task.
        for task_ in dag.tasks:
            task_.update_envs({'SKYPILOT_NUM_JOBS': str(num_jobs)})

        dag_utils.dump_dag_to_yaml(dag, f.name)

        vars_to_fill: Dict[str, Any] = {
            'remote_original_user_yaml_path': remote_orig_user_yaml_path,
            'original_user_dag_path': original_user_yaml_path.name,
            'remote_user_yaml_path': remote_user_yaml_path,
            'user_yaml_path': f.name,
            'local_to_controller_file_mounts':
                (local_to_controller_file_mounts),
            'jobs_controller': controller_name,
            'dag_name': dag.name,
            'remote_user_config_path': remote_user_config_path,
            'remote_env_file_path': remote_env_file_path,
            'modified_catalogs': modified_catalogs,
            'priority': priority,
            'is_consolidation_mode': is_consolidation_mode,
            'pool': pool,
            'job_controller_indicator_file':
                managed_job_constants.JOB_CONTROLLER_INDICATOR_FILE,
            'num_jobs': num_jobs,
            **controller_utils.shared_controller_vars_to_fill(
                controller,
                remote_user_config_path=remote_user_config_path,
                # TODO(aylei): the mutated config will not be updated
                # afterwards without recreate the controller. Need to
                # revisit this.
                local_user_config=mutated_user_config,
            ),
        }

        yaml_path = os.path.join(
            managed_job_constants.JOBS_CONTROLLER_YAML_PREFIX,
            f'{name}-{dag_uuid}.yaml')

        # Launch with the api server's user hash, so that sky status does
        # not show the owner of the controller as whatever user launched
        # it first.
        with common.with_server_user():
            # Always launch the controller in the default workspace.
            with skypilot_config.local_active_workspace_ctx(
                    skylet_constants.SKYPILOT_DEFAULT_WORKSPACE):
                job_controller_postfix = (' from jobs controller'
                                          if not is_consolidation_mode else '')
                managed_jobs_str = 'managed job'

                job_ids_str = _job_ids_to_str(job_ids)
                vars_to_fill['job_ids'] = job_ids
                # Create job_id_to_rank dictionary by sorting job IDs and
                # assigning ranks.
                sorted_job_ids = sorted(job_ids)
                job_id_to_rank = {
                    str(job_id): rank
                    for rank, job_id in enumerate(sorted_job_ids)
                }
                vars_to_fill['job_id_to_rank'] = job_id_to_rank
                if num_jobs is not None and num_jobs > 1:
                    managed_jobs_str = (
                        f'{num_jobs} managed jobs {job_ids_str}')
                logger.info(
                    f'{colorama.Fore.YELLOW}'
                    f'Launching {managed_jobs_str} {dag.name!r}'
                    f'{job_controller_postfix}...{colorama.Style.RESET_ALL}')

                common_utils.fill_template(
                    managed_job_constants.JOBS_CONTROLLER_TEMPLATE,
                    vars_to_fill,
                    output_path=yaml_path)
                logger.debug(f'Wrote controller yaml to path: {yaml_path}')
                controller_task = task_lib.Task.from_yaml(yaml_path)
                controller_task.set_resources(controller_resources)
                controller_task.managed_job_dag = dag
                # pylint: disable=protected-access
                controller_task._metadata = metadata

                # TODO(zhwu): the buckets need to be correctly handled for
                # a specific workspace. For example, if a job is launched in
                # workspace A, but the controller is in workspace B, the
                # intermediate bucket and newly created bucket should be in
                # workspace A.
                if is_consolidation_mode:
                    return _consolidated_launch(controller, controller_task,
                                                job_ids)
                else:
                    # TODO(lloyd-brown) The cluster should already be launched
                    # here so we should just be able to use exec, but we need
                    # to work through the logic and make sure there is no issue
                    # with say file mounts.

                    # Job controller is not placed in kueue, as the
                    # controller pod is considered a "system" pod
                    # and is not subject to queue limits or preemption.
                    with skypilot_config.remove_queue_name_from_config():
                        result = execution.launch(
                            task=controller_task,
                            cluster_name=controller_name,
                            stream_logs=stream_logs,
                            retry_until_up=True,
                            fast=True,
                            _request_name=request_names.AdminPolicyRequestName.
                            JOBS_LAUNCH_CONTROLLER,
                            _disable_controller_check=True)
                        return job_ids, result[1]


def queue_from_kubernetes_pod(
        pod_name: str,
        context: Optional[str] = None,
        skip_finished: bool = False) -> List[Dict[str, Any]]:
    """Gets the jobs queue from a specific controller pod.

    Args:
        pod_name (str): The name of the controller pod to query for jobs.
        context (Optional[str]): The Kubernetes context to use. If None, the
            current context is used.
        skip_finished (bool): If True, does not return finished jobs.

    Returns:
        [
            {
                'job_id': int,
                'job_name': str,
                'resources': str,
                'submitted_at': (float) timestamp of submission,
                'end_at': (float) timestamp of end,
                'duration': (float) duration in seconds,
                'recovery_count': (int) Number of retries,
                'status': (sky.jobs.ManagedJobStatus) of the job,
                'cluster_resources': (str) resources of the cluster,
                'region': (str) region of the cluster,
            }
        ]

    Raises:
        RuntimeError: If there's an error fetching the managed jobs.
    """
    # Create dummy cluster info to get the command runner.
    provider_config = {'context': context}
    instances = {
        pod_name: [
            provision_common.InstanceInfo(instance_id=pod_name,
                                          internal_ip='',
                                          external_ip='',
                                          tags={})
        ]
    }  # Internal IP is not required for Kubernetes
    cluster_info = provision_common.ClusterInfo(provider_name='kubernetes',
                                                head_instance_id=pod_name,
                                                provider_config=provider_config,
                                                instances=instances)
    managed_jobs_runner = provision_lib.get_command_runners(
        'kubernetes', cluster_info)[0]

    code = managed_job_utils.ManagedJobCodeGen.get_job_table(
        skip_finished=skip_finished,
        fields=_MANAGED_JOB_FIELDS_FOR_QUEUE_KUBERNETES)
    returncode, job_table_payload, stderr = managed_jobs_runner.run(
        code,
        require_outputs=True,
        separate_stderr=True,
        stream_logs=False,
    )
    try:
        subprocess_utils.handle_returncode(returncode,
                                           code,
                                           'Failed to fetch managed jobs',
                                           job_table_payload + stderr,
                                           stream_logs=False)
    except exceptions.CommandError as e:
        raise RuntimeError(str(e)) from e

    jobs, _, result_type, _, _ = managed_job_utils.load_managed_job_queue(
        job_table_payload)

    if result_type == managed_job_utils.ManagedJobQueueResultType.DICT:
        return jobs

    # Backward compatibility for old jobs controller without filtering
    # TODO(hailong): remove this after 0.12.0
    if skip_finished:
        # Filter out the finished jobs. If a multi-task job is partially
        # finished, we will include all its tasks.
        non_finished_tasks = list(
            filter(lambda job: not job['status'].is_terminal(), jobs))
        non_finished_job_ids = {job['job_id'] for job in non_finished_tasks}
        jobs = list(
            filter(lambda job: job['job_id'] in non_finished_job_ids, jobs))
    return jobs


def _maybe_restart_controller(
        refresh: bool, stopped_message: str, spinner_message: str
) -> 'cloud_vm_ray_backend.CloudVmRayResourceHandle':
    """Restart controller if refresh is True and it is stopped."""
    jobs_controller_type = controller_utils.Controllers.JOBS_CONTROLLER
    if refresh:
        stopped_message = ''
    try:
        handle = backend_utils.is_controller_accessible(
            controller=jobs_controller_type, stopped_message=stopped_message)
    except exceptions.ClusterNotUpError as e:
        if not refresh:
            raise
        handle = None
        controller_status = e.cluster_status

    if handle is not None:
        return handle

    logger.info(f'{colorama.Fore.YELLOW}'
                f'Restarting {jobs_controller_type.value.name}...'
                f'{colorama.Style.RESET_ALL}')

    rich_utils.force_update_status(
        ux_utils.spinner_message(f'{spinner_message} - restarting '
                                 'controller'))
    with skypilot_config.local_active_workspace_ctx(
            skylet_constants.SKYPILOT_DEFAULT_WORKSPACE):
        global_user_state.add_cluster_event(
            jobs_controller_type.value.cluster_name,
            status_lib.ClusterStatus.INIT, 'Jobs controller restarted.',
            global_user_state.ClusterEventType.STATUS_CHANGE)
        handle = core.start(
            cluster_name=jobs_controller_type.value.cluster_name)

    controller_status = status_lib.ClusterStatus.UP
    rich_utils.force_update_status(ux_utils.spinner_message(spinner_message))

    assert handle is not None, (controller_status, refresh)
    return handle


# For backwards compatibility
# TODO(hailong): Remove before 0.12.0.
@usage_lib.entrypoint
def queue(refresh: bool,
          skip_finished: bool = False,
          all_users: bool = False,
          job_ids: Optional[List[int]] = None) -> List[Dict[str, Any]]:
    # NOTE(dev): Keep the docstring consistent between the Python API and CLI.
    """Gets statuses of managed jobs.

    Please refer to sky.cli.job_queue for documentation.

    Returns:
        [
            {
                'job_id': int,
                'job_name': str,
                'resources': str,
                'submitted_at': (float) timestamp of submission,
                'end_at': (float) timestamp of end,
                'job_duration': (float) duration in seconds,
                'recovery_count': (int) Number of retries,
                'status': (sky.jobs.ManagedJobStatus) of the job,
                'cluster_resources': (str) resources of the cluster,
                'region': (str) region of the cluster,
                'user_name': (Optional[str]) job creator's user name,
                'user_hash': (str) job creator's user hash,
                'task_id': (int), set to 0 (except in pipelines, which may have multiple tasks), # pylint: disable=line-too-long
                'task_name': (str), same as job_name (except in pipelines, which may have multiple tasks), # pylint: disable=line-too-long
            }
        ]
    Raises:
        sky.exceptions.ClusterNotUpError: the jobs controller is not up or
            does not exist.
        RuntimeError: if failed to get the managed jobs with ssh.
    """
    jobs, _, _, _ = queue_v2(refresh, skip_finished, all_users, job_ids)

    return jobs


@usage_lib.entrypoint
def queue_v2_api(
    refresh: bool,
    skip_finished: bool = False,
    all_users: bool = False,
    job_ids: Optional[List[int]] = None,
    user_match: Optional[str] = None,
    workspace_match: Optional[str] = None,
    name_match: Optional[str] = None,
    pool_match: Optional[str] = None,
    page: Optional[int] = None,
    limit: Optional[int] = None,
    statuses: Optional[List[str]] = None,
    fields: Optional[List[str]] = None,
    sort_by: Optional[str] = None,
    sort_order: Optional[str] = None,
) -> Tuple[List[responses.ManagedJobRecord], int, Dict[str, int], int]:
    """Gets statuses of managed jobs and parse the
    jobs to responses.ManagedJobRecord."""
    jobs, total, status_counts, total_no_filter = queue_v2(
        refresh, skip_finished, all_users, job_ids, user_match, workspace_match,
        name_match, pool_match, page, limit, statuses, fields, sort_by,
        sort_order)
    return [responses.ManagedJobRecord(**job) for job in jobs
           ], total, status_counts, total_no_filter


@metrics_lib.time_me
def queue_v2(
    refresh: bool,
    skip_finished: bool = False,
    all_users: bool = False,
    job_ids: Optional[List[int]] = None,
    user_match: Optional[str] = None,
    workspace_match: Optional[str] = None,
    name_match: Optional[str] = None,
    pool_match: Optional[str] = None,
    page: Optional[int] = None,
    limit: Optional[int] = None,
    statuses: Optional[List[str]] = None,
    fields: Optional[List[str]] = None,
    sort_by: Optional[str] = None,
    sort_order: Optional[str] = None,
) -> Tuple[List[Dict[str, Any]], int, Dict[str, int], int]:
    # NOTE(dev): Keep the docstring consistent between the Python API and CLI.
    """Gets statuses of managed jobs with filtering.

    Please refer to sky.cli.job_queue for documentation.

    Returns:
        jobs: List[Dict[str, Any]]
            [
                {
                    'job_id': int,
                    'job_name': str,
                    'resources': str,
                    'submitted_at': (float) timestamp of submission,
                    'end_at': (float) timestamp of end,
                    'job_duration': (float) duration in seconds,
                    'recovery_count': (int) Number of retries,
                    'status': (sky.jobs.ManagedJobStatus) of the job,
                    'cluster_resources': (str) resources of the cluster,
                    'region': (str) region of the cluster,
                    'user_name': (Optional[str]) job creator's user name,
                    'user_hash': (str) job creator's user hash,
                    'task_id': (int), set to 0 (except in pipelines, which may have multiple tasks), # pylint: disable=line-too-long
                    'task_name': (str), same as job_name (except in pipelines, which may have multiple tasks), # pylint: disable=line-too-long
                }
            ]
        total: int, total number of jobs after filter
        status_counts: Dict[str, int], status counts after filter
        total_no_filter: int, total number of jobs before filter
    Raises:
        sky.exceptions.ClusterNotUpError: the jobs controller is not up or
            does not exist.
        RuntimeError: if failed to get the managed jobs with ssh.
    """
    if limit is not None:
        if limit < 1:
            raise ValueError(f'Limit must be at least 1, got {limit}')
        if page is None:
            page = 1
        if page < 1:
            raise ValueError(f'Page must be at least 1, got {page}')
    else:
        if page is not None:
            raise ValueError('Limit must be specified when page is specified')

    with metrics_lib.time_it('jobs.queue.restart_controller', group='jobs'):
        handle = _maybe_restart_controller(refresh,
                                           stopped_message='No in-progress '
                                           'managed jobs.',
                                           spinner_message='Checking '
                                           'managed jobs')
    backend = backend_utils.get_backend_from_handle(handle)
    assert isinstance(backend, backends.CloudVmRayBackend)

    user_hashes: Optional[List[Optional[str]]] = None
    show_jobs_without_user_hash = False
    if not all_users:
        user_hashes = [common_utils.get_user_hash()]
        # For backwards compatibility, we show jobs that do not have a
        # user_hash. TODO(cooperc): Remove before 0.12.0.
        user_hashes.append(None)
        show_jobs_without_user_hash = True
    elif user_match is not None:
        users = global_user_state.get_user_by_name_match(user_match)
        if not users:
            return [], 0, {}, 0
        user_hashes = [user.id for user in users]

    accessible_workspaces = list(workspaces_core.get_workspaces().keys())

    if handle.is_grpc_enabled_with_flag:
        try:
            request = managed_jobsv1_pb2.GetJobTableRequest(
                skip_finished=skip_finished,
                accessible_workspaces=(managed_jobsv1_pb2.Workspaces(
                    workspaces=accessible_workspaces)),
                job_ids=managed_jobsv1_pb2.JobIds(
                    ids=job_ids) if job_ids is not None else None,
                workspace_match=workspace_match,
                name_match=name_match,
                pool_match=pool_match,
                page=page,
                limit=limit,
                # Remove None from user_hashes, as the gRPC server uses the
                # show_jobs_without_user_hash flag instead.
                user_hashes=managed_jobsv1_pb2.UserHashes(hashes=[
                    user_hash for user_hash in user_hashes
                    if user_hash is not None
                ]) if user_hashes is not None else None,
                statuses=managed_jobsv1_pb2.Statuses(
                    statuses=statuses) if statuses is not None else None,
                fields=managed_jobsv1_pb2.Fields(
                    fields=fields) if fields is not None else None,
                show_jobs_without_user_hash=show_jobs_without_user_hash,
                sort_by=sort_by,
                sort_order=sort_order,
            )
            response = backend_utils.invoke_skylet_with_retries(
                lambda: cloud_vm_ray_backend.SkyletClient(
                    handle.get_grpc_channel()).get_managed_job_table(request))
            jobs = managed_job_utils.decode_managed_job_protos(response.jobs)
            return jobs, response.total, dict(
                response.status_counts), response.total_no_filter
        except exceptions.SkyletMethodNotImplementedError:
            pass

    with metrics_lib.time_it('jobs.queue.generate_code', group='jobs'):
        code = managed_job_utils.ManagedJobCodeGen.get_job_table(
            skip_finished, accessible_workspaces, job_ids, workspace_match,
            name_match, pool_match, page, limit, user_hashes, statuses, fields,
            sort_by, sort_order)
    with metrics_lib.time_it('jobs.queue.run_on_head', group='jobs'):
        returncode, job_table_payload, stderr = backend.run_on_head(
            handle,
            code,
            require_outputs=True,
            stream_logs=False,
            separate_stderr=True)

    if returncode != 0:
        logger.error(job_table_payload + stderr)
        raise RuntimeError('Failed to fetch managed jobs with returncode: '
                           f'{returncode}.\n{job_table_payload + stderr}')

    with metrics_lib.time_it('jobs.queue.load_job_queue', group='jobs'):
        (jobs, total, result_type, total_no_filter, status_counts
        ) = managed_job_utils.load_managed_job_queue(job_table_payload)

    if result_type == managed_job_utils.ManagedJobQueueResultType.DICT:
        return jobs, total, status_counts, total_no_filter

    # Backward compatibility for old jobs controller without filtering
    # TODO(hailong): remove this after 0.12.0
    with metrics_lib.time_it('jobs.queue.filter_and_process', group='jobs'):
        if not all_users:

            def user_hash_matches_or_missing(job: Dict[str, Any]) -> bool:
                user_hash = job.get('user_hash', None)
                if user_hash is None:
                    # For backwards compatibility, we show jobs that do not have
                    # a user_hash. TODO(cooperc): Remove before 0.12.0.
                    return True
                return user_hash == common_utils.get_user_hash()

            jobs = list(filter(user_hash_matches_or_missing, jobs))

        jobs = list(
            filter(
                lambda job: job.get('workspace', skylet_constants.
                                    SKYPILOT_DEFAULT_WORKSPACE) in
                accessible_workspaces, jobs))

        if skip_finished:
            # Filter out the finished jobs. If a multi-task job is partially
            # finished, we will include all its tasks.
            non_finished_tasks = list(
                filter(lambda job: not job['status'].is_terminal(), jobs))
            non_finished_job_ids = {job['job_id'] for job in non_finished_tasks}
            jobs = list(
                filter(lambda job: job['job_id'] in non_finished_job_ids, jobs))

        if job_ids:
            jobs = [job for job in jobs if job['job_id'] in job_ids]

        filtered_jobs, total, status_counts = managed_job_utils.filter_jobs(
            jobs,
            workspace_match,
            name_match,
            pool_match,
            page=page,
            limit=limit,
            user_match=user_match,
            enable_user_match=True,
            statuses=statuses,
        )
    return filtered_jobs, total, status_counts, total_no_filter


@usage_lib.entrypoint
# pylint: disable=redefined-builtin
def cancel(name: Optional[str] = None,
           job_ids: Optional[List[int]] = None,
           all: bool = False,
           all_users: bool = False,
           pool: Optional[str] = None,
           graceful: bool = False,
           graceful_timeout: Optional[int] = None) -> None:
    # NOTE(dev): Keep the docstring consistent between the Python API and CLI.
    """Cancels managed jobs.

    Please refer to sky.cli.job_cancel for documentation.

    Raises:
        sky.exceptions.ClusterNotUpError: the jobs controller is not up.
        RuntimeError: failed to cancel the job.
    """
    with rich_utils.safe_status(
            ux_utils.spinner_message('Cancelling managed jobs')):
        job_ids = [] if job_ids is None else job_ids
        handle = backend_utils.is_controller_accessible(
            controller=controller_utils.Controllers.JOBS_CONTROLLER,
            stopped_message='All managed jobs should have finished.')

        job_id_str = ','.join(map(str, job_ids))
        if sum([
                bool(job_ids), name is not None, pool is not None, all or
                all_users
        ]) != 1:
            arguments = []
            arguments += [f'job_ids={job_id_str}'] if job_ids else []
            arguments += [f'name={name}'] if name is not None else []
            arguments += [f'pool={pool}'] if pool is not None else []
            arguments += ['all'] if all else []
            arguments += ['all_users'] if all_users else []
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    'Can only specify one of JOB_IDS, name, pool, or all/'
                    f'all_users. Provided {" ".join(arguments)!r}.')

        job_ids = None if (all_users or all) else job_ids

        backend = backend_utils.get_backend_from_handle(handle)
        assert isinstance(backend, backends.CloudVmRayBackend)

        use_legacy = not handle.is_grpc_enabled_with_flag

        if not use_legacy:
            current_workspace = skypilot_config.get_active_workspace()
            try:
                request = managed_jobsv1_pb2.CancelJobsRequest(
                    current_workspace=current_workspace,
                    graceful=graceful,
                    graceful_timeout=graceful_timeout)

                if all_users or all or job_ids:
                    request.all_users = all_users
                    if all:
                        request.user_hash = common_utils.get_user_hash()
                    if job_ids is not None:
                        request.job_ids.CopyFrom(
                            managed_jobsv1_pb2.JobIds(ids=job_ids))
                elif name is not None:
                    request.job_name = name
                else:
                    assert pool is not None, (job_ids, name, pool, all)
                    request.pool_name = pool

                response = backend_utils.invoke_skylet_with_retries(
                    lambda: cloud_vm_ray_backend.SkyletClient(
                        handle.get_grpc_channel()).cancel_managed_jobs(request))
                stdout = response.message
            except exceptions.SkyletMethodNotImplementedError:
                use_legacy = True

        if use_legacy:
            if all_users or all or job_ids:
                code = managed_job_utils.ManagedJobCodeGen.cancel_jobs_by_id(
                    job_ids,
                    all_users=all_users,
                    graceful=graceful,
                    graceful_timeout=graceful_timeout)
            elif name is not None:
                code = managed_job_utils.ManagedJobCodeGen.cancel_job_by_name(
                    name, graceful=graceful, graceful_timeout=graceful_timeout)
            else:
                assert pool is not None, (job_ids, name, pool, all)
                code = managed_job_utils.ManagedJobCodeGen.cancel_jobs_by_pool(
                    pool)
            # The stderr is redirected to stdout
            returncode, stdout, stderr = backend.run_on_head(
                handle, code, require_outputs=True, stream_logs=False)
            try:
                subprocess_utils.handle_returncode(
                    returncode, code, 'Failed to cancel managed job',
                    stdout + stderr)
            except exceptions.CommandError as e:
                with ux_utils.print_exception_no_traceback():
                    raise RuntimeError(e.error_msg) from e

        logger.info(stdout)
        if 'Multiple jobs found with name' in stdout:
            with ux_utils.print_exception_no_traceback():
                raise RuntimeError(
                    'Please specify the job ID instead of the job name.')


@usage_lib.entrypoint
def tail_logs(name: Optional[str],
              job_id: Optional[int],
              follow: bool,
              controller: bool,
              refresh: bool,
              tail: Optional[int] = None,
              task: Optional[Union[str, int]] = None) -> int:
    # NOTE(dev): Keep the docstring consistent between the Python API and CLI.
    """Tail logs of managed jobs.

    Please refer to sky.cli.job_logs for documentation.

    Returns:
        Exit code based on success or failure of the job. 0 if success,
        100 if the job failed. See exceptions.JobExitCode for possible exit
        codes.

    Raises:
        ValueError: invalid arguments.
        sky.exceptions.ClusterNotUpError: the jobs controller is not up.
    """
    # TODO(zhwu): Automatically restart the jobs controller
    if name is not None and job_id is not None:
        with ux_utils.print_exception_no_traceback():
            raise ValueError('Cannot specify both name and job_id.')

    jobs_controller_type = controller_utils.Controllers.JOBS_CONTROLLER
    job_name_or_id_str = ''
    if job_id is not None:
        job_name_or_id_str = str(job_id)
    elif name is not None:
        job_name_or_id_str = f'-n {name}'
    else:
        job_name_or_id_str = ''
    handle = _maybe_restart_controller(
        refresh,
        stopped_message=(
            f'{jobs_controller_type.value.name.capitalize()} is stopped. To '
            f'get the logs, run: {colorama.Style.BRIGHT}sky jobs logs '
            f'-r {job_name_or_id_str}{colorama.Style.RESET_ALL}'),
        spinner_message='Retrieving job logs')

    backend = backend_utils.get_backend_from_handle(handle)
    assert isinstance(backend, backends.CloudVmRayBackend), backend

    return backend.tail_managed_job_logs(handle,
                                         job_id=job_id,
                                         job_name=name,
                                         follow=follow,
                                         controller=controller,
                                         tail=tail,
                                         task=task)


@usage_lib.entrypoint
def download_logs(
        name: Optional[str],
        job_id: Optional[int],
        refresh: bool,
        controller: bool,
        local_dir: str = skylet_constants.SKY_LOGS_DIRECTORY) -> Dict[str, str]:
    """Sync down logs of managed jobs.

    Please refer to sky.cli.job_logs for documentation.

    Returns:
        A dictionary mapping job ID to the local path.

    Raises:
        ValueError: invalid arguments.
        sky.exceptions.ClusterNotUpError: the jobs controller is not up.
    """
    if name is not None and job_id is not None:
        with ux_utils.print_exception_no_traceback():
            raise ValueError('Cannot specify both name and job_id.')

    jobs_controller_type = controller_utils.Controllers.JOBS_CONTROLLER
    job_name_or_id_str = ''
    if job_id is not None:
        job_name_or_id_str = str(job_id)
    elif name is not None:
        job_name_or_id_str = f'-n {name}'
    else:
        job_name_or_id_str = ''
    handle = _maybe_restart_controller(
        refresh,
        stopped_message=(
            f'{jobs_controller_type.value.name.capitalize()} is stopped. To '
            f'get the logs, run: {colorama.Style.BRIGHT}sky jobs logs '
            f'-r --sync-down {job_name_or_id_str}{colorama.Style.RESET_ALL}'),
        spinner_message='Retrieving job logs')

    backend = backend_utils.get_backend_from_handle(handle)
    assert isinstance(backend, backends.CloudVmRayBackend), backend

    return backend.sync_down_managed_job_logs(handle,
                                              job_id=job_id,
                                              job_name=name,
                                              controller=controller,
                                              local_dir=local_dir)


@usage_lib.entrypoint
def pool_apply(
    task: 'sky.Task',
    pool_name: str,
    mode: serve_utils.UpdateMode = serve_utils.DEFAULT_UPDATE_MODE,
    workers: Optional[int] = None,
) -> None:
    """Apply a config to a pool."""
    return impl.apply(task, workers, pool_name, mode, pool=True)


@usage_lib.entrypoint
# pylint: disable=redefined-builtin
def pool_down(
    pool_names: Optional[Union[str, List[str]]] = None,
    all: bool = False,
    purge: bool = False,
) -> None:
    """Delete a pool."""
    return impl.down(pool_names, all, purge, pool=True)


@usage_lib.entrypoint
def pool_status(
    pool_names: Optional[Union[str,
                               List[str]]] = None,) -> List[Dict[str, Any]]:
    """Query a pool."""
    return impl.status(pool_names, pool=True)


ServiceComponentOrStr = Union[str, serve_utils.ServiceComponent]


@usage_lib.entrypoint
def pool_tail_logs(
    pool_name: str,
    *,
    target: ServiceComponentOrStr,
    worker_id: Optional[int] = None,
    follow: bool = True,
    tail: Optional[int] = None,
) -> None:
    """Tail logs of a pool."""
    return impl.tail_logs(pool_name,
                          target=target,
                          replica_id=worker_id,
                          follow=follow,
                          tail=tail,
                          pool=True)


@usage_lib.entrypoint
def pool_sync_down_logs(
    pool_name: str,
    *,
    local_dir: str,
    targets: Union[ServiceComponentOrStr, List[ServiceComponentOrStr],
                   None] = None,
    worker_ids: Optional[List[int]] = None,
    tail: Optional[int] = None,
) -> str:
    """Sync down logs of a pool."""
    return impl.sync_down_logs(pool_name,
                               local_dir=local_dir,
                               targets=targets,
                               replica_ids=worker_ids,
                               tail=tail,
                               pool=True)


@usage_lib.entrypoint
def get_job_events(
    job_id: int,
    task_id: Optional[int] = None,
    limit: Optional[int] = 10,
) -> List[Dict[str, Any]]:
    """Get task events for a managed job.

    Args:
        job_id: The job ID to get task events for.
        task_id: Optional task ID to filter by.
        limit: Optional limit on number of task events to return (default 10).

    Returns:
        List of task event records.
    """
    return managed_job_state.get_job_events(job_id=job_id,
                                            task_id=task_id,
                                            limit=limit)



"""
module allocating VPC - network namespace and configuration
for Ray's cluster. used by the node_provider module to group the
nodes under the same subnet, tagged by the same cluster name.
"""

import copy
import json
import textwrap
import time
import uuid
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Dict

import requests

from sky.adaptors import ibm
from sky.skylet.providers.ibm.utils import RAY_RECYCLABLE, get_logger

# pylint: disable=line-too-long
logger = get_logger("vpc_provider_")
REQUIRED_RULES = {
    "outbound_tcp_all": "selected security group is missing rule permitting outbound TCP access\n",
    "outbound_udp_all": "selected security group is missing rule permitting outbound UDP access\n",
    "inbound_tcp_sg": "selected security group is missing rule permitting inbound tcp traffic inside selected security group\n",
    "inbound_tcp_22": "selected security group is missing rule permitting inbound traffic to tcp port 22 required for ssh\n",
}
INSECURE_RULES = {
    "inbound_tcp_6379": "selected security group is missing rule permitting inbound traffic to tcp port 6379 required for Redis\n",
    "inbound_tcp_8265": "selected security group is missing rule permitting inbound traffic to tcp port 8265 required to access Ray Dashboard\n",
}
REQUIRED_RULES.update(INSECURE_RULES)


class IBMVPCProvider:
    """
    Manages a vpc containing the network configuration for the cluster's nodes.
    """

    def __init__(self, resource_group_id, region, cluster_name):
        self.vpc_client = ibm.client(region=region)
        self.search_client = ibm.search_client()
        self.tagging_client = ibm.tagging_client()
        self.resource_group_id = resource_group_id
        self.cluster_name = cluster_name
        ## region and zone might change between failovers
        self.region = None
        self.zone = None

    def create_or_fetch_vpc(self, region, zone):
        """
        returns a cluster with tag matching the cluster name if exists, else creates one.
        an entry point (out of 2) to this module.
        """

        # refresh client region scope if region changed.
        if self.region and self.region != region:
            self.vpc_client = ibm.client(region=region)
        self.region = region
        self.zone = zone
        reused_vpc_data = None
        # pylint: disable=line-too-long
        vpcs_filtered_by_tags_and_region = self.search_client.search(
            query=f"type:vpc AND tags:{self.cluster_name} AND region:{self.region}",
            fields=["tags", "region", "type"],
            limit=1000,
        ).get_result()["items"]
        for vpc in vpcs_filtered_by_tags_and_region:
            vpc_id = vpc["crn"].rsplit(":", 1)[-1]
            vpc_data = self.get_vpc_data(vpc_id, self.region)
            if vpc_data["status"] == "available":
                reused_vpc_data = vpc_data
                break
        # found vpc tagged with cluster name in the required region
        if reused_vpc_data:
            # using self.region since tagged vpc is in the same region
            subnets = self.get_vpc_subnets(reused_vpc_data, self.region)
            subnet_in_zone = next(
                (subnet for subnet in subnets if subnet["zone"]["name"] == self.zone),
                None,
            )
            # found a subnet in the required zone
            if subnet_in_zone:
                subnet_id = subnet_in_zone["id"]
                public_gateway = subnet_in_zone.get("public_gateway")
                if not public_gateway:
                    public_gateway = self.create_public_gateway(
                        reused_vpc_data["id"], self.zone, subnet_in_zone
                    )
            # tagged vpc found doesn't have a subnet in the required zone
            else:
                subnet_data = self.create_subnet(reused_vpc_data["id"], self.zone)
                subnet_id = subnet_data["id"]
                public_gateway = self.create_public_gateway(
                    reused_vpc_data["id"], self.zone, subnet_data
                )

            # add missing security group rules if needed
            security_group = reused_vpc_data.get("default_security_group")
            if security_group:
                sg_id = security_group["id"]
                self.add_missing_sg_rules(sg_id)

                # managed to reuse found VPC
                logger.info(
                    f"Reusing VPC {reused_vpc_data['id']} named: {reused_vpc_data['name']}"
                )
                return {
                    "vpc_id": reused_vpc_data["id"],
                    "subnet_id": subnet_id,
                    "security_group_id": sg_id,
                }

        # delete a tagged vpc that doesn't meet requirements
        if reused_vpc_data:
            self.delete_vpc(reused_vpc_data["id"], self.region)
        # create a new vpc
        vpc_tags = self.create_vpc()
        return vpc_tags

    def create_vpc(self):
        """creates a vpc, tags it and return key values pertaining it.

        uuid is added to vpc name to avoid naming collision.
        vpc is tagged using the exact cluster name, as appears on
            the cluster's config file.

        Returns:
            dict: containing the keys: vpc_id, subnet_id
                    and security_group_id
        """
        vpc_data = self.vpc_client.create_vpc(
            address_prefix_management="auto",
            classic_access=False,
            name=f"sky-vpc-{self.cluster_name}-{str(uuid.uuid4())[:5]}",
            resource_group={"id": self.resource_group_id},
        ).get_result()
        subnet_data = self.create_subnet(vpc_data["id"], self.zone)
        self.create_public_gateway(vpc_data["id"], self.zone, subnet_data)
        sg_id = self.create_sg_rules(vpc_data)

        # tag vpc with the cluster's name
        resource_model = {"resource_id": vpc_data["crn"]}
        self.tagging_client.attach_tag(
            resources=[resource_model], tag_names=[self.cluster_name], tag_type="user"
        ).get_result()

        return {
            "vpc_id": vpc_data["id"],
            "subnet_id": subnet_data["id"],
            "security_group_id": sg_id,
        }

    def create_subnet(self, vpc_id, zone_name):
        ipv4_cidr_block = None
        subnet_name = f"sky-subnet-{self.cluster_name}-{str(uuid.uuid4())[:5]}"
        res = self.vpc_client.list_vpc_address_prefixes(vpc_id).get_result()

        # searching for the CIDR block (internal ip range) matching the
        # specified zone of a VPC (whose region has already been set)
        address_prefixes = res["address_prefixes"]
        ipv4_cidr_block = next(
            (
                address_prefix["cidr"]
                for address_prefix in address_prefixes
                if address_prefix["zone"]["name"] == zone_name
            ),
            None,
        )
        if not ipv4_cidr_block:
            raise Exception(
                "Failed to locate a cidr block "
                f"Matching the zone name: {zone_name} to create "
                "a subnet"
            )

        subnet_prototype: Dict[str, Any] = {}
        subnet_prototype["zone"] = {"name": zone_name}
        subnet_prototype["ip_version"] = "ipv4"
        subnet_prototype["name"] = subnet_name
        subnet_prototype["resource_group"] = {"id": self.resource_group_id}
        subnet_prototype["vpc"] = {"id": vpc_id}
        subnet_prototype["ipv4_cidr_block"] = ipv4_cidr_block

        subnet_data = self.vpc_client.create_subnet(subnet_prototype).get_result()
        return subnet_data

    def create_public_gateway(self, vpc_id, zone_name, subnet_data):

        gateway_prototype: Dict[str, Any] = {}
        gateway_prototype["vpc"] = {"id": vpc_id}
        gateway_prototype["zone"] = {"name": zone_name}
        gateway_prototype["name"] = f"{subnet_data['name']}-gw"
        gateway_prototype["resource_group"] = {"id": self.resource_group_id}
        gateway_data = self.vpc_client.create_public_gateway(
            **gateway_prototype
        ).get_result()
        gateway_id = gateway_data["id"]

        self.vpc_client.set_subnet_public_gateway(subnet_data["id"], {"id": gateway_id})
        return gateway_id

    def create_sg_rules(self, vpc_data):

        sg_id = vpc_data["default_security_group"]["id"]
        sg_name = f"{self.cluster_name}-{str(uuid.uuid4())[:5]}-sg"

        # update sg name
        self.vpc_client.update_security_group(
            sg_id, security_group_patch={"name": sg_name}
        )

        # open private tcp traffic between VSIs within the security group
        sg_rule_prototype = _build_security_group_rule_prototype_model(
            "inbound_tcp_sg", sg_id=sg_id
        )
        self.vpc_client.create_security_group_rule(
            sg_id, sg_rule_prototype
        ).get_result()

        # add all other required rules configured by the specific backend
        for rule in REQUIRED_RULES.keys():
            sg_rule_prototype = _build_security_group_rule_prototype_model(rule)
            if sg_rule_prototype:
                self.vpc_client.create_security_group_rule(
                    sg_id, sg_rule_prototype
                ).get_result()

        return sg_id

    def get_vpc_data(self, vpc_id, region):
        """returns vpc data if exists, else None"""
        if not vpc_id:
            return None
        tmp_vpc_client = ibm.client(region=region)
        try:
            vpc_data = tmp_vpc_client.get_vpc(vpc_id).result
            return vpc_data
        except ibm.ibm_cloud_sdk_core.ApiException as e:
            if e.code == 404:
                logger.debug("VPC doesn't exist.")
                return None
            else:
                raise

    def get_vpc_subnets(self, vpc_data, region, field=""):
        """return data on subnets belonging to specified vpc within
        the specified region.

        if 'field' is specified narrowing data returned to
        data['field'] (for each subnet)

        Args:
            vpc_data (str): vpc data as received from vpc client.
            region (str): ibm vpc region.
            field (str, optional): field within the subnet data response from vpc
              client's list_subnets()/get_subnet() response. Defaults to ''.

        Returns:
            str: data on all subnets belonging to the specified vpc.
        """
        if not vpc_data:
            return None
        # pylint: disable=line-too-long
        tmp_vpc_client = ibm.client(region=region)
        subnets_attached_to_routing_table = tmp_vpc_client.list_subnets(
            routing_table_id=vpc_data["default_routing_table"]["id"]
        ).get_result()["subnets"]
        if field:
            return [subnet[field] for subnet in subnets_attached_to_routing_table]
        else:
            return subnets_attached_to_routing_table

    def delete_vpc(self, vpc_id, region):
        """
        deletes a vpc with the specified id and region.
        an entry point to this module (alongside create_or_fetch_vpc)
        """
        logger.debug(f"Deleting vpc: {vpc_id}")
        tmp_vpc_client = ibm.client(region=region)
        vpc_data = self.get_vpc_data(vpc_id, region)
        if not vpc_data:
            logger.warn(f"vpc:{vpc_id} is set for deletion, but wasn't found")
            return None
        self.delete_vms(tmp_vpc_client, vpc_id)
        self.delete_subnets(tmp_vpc_client, vpc_data, region)
        self.delete_gateways(tmp_vpc_client, vpc_id)
        # at this point vpc was already verified to be existing
        # thus no relevant exception to catch when deleting.
        tmp_vpc_client.delete_vpc(vpc_id)

    def delete_vms(self, vpc_client, vpc_id):
        def _poll_vpc_contains_vms(vpc_id):
            tries = 60
            sleep_interval = 3
            while tries:
                # list_instances() never raise an exception, check values instead
                res = vpc_client.list_instances(vpc_id=vpc_id).get_result()
                if not res["total_count"]:
                    return True
                else:
                    tries -= 1
                    time.sleep(sleep_interval)
            raise Exception(
                "Failed to delete VPC's instances within "
                "the expected time frame. Cannot "
                "continue to delete VPC."
            )

        def _del_instance(vm_data):
            # first delete ips created by node_provider
            nic_id = vm_data["network_interfaces"][0]["id"]
            res = vpc_client.list_instance_network_interface_floating_ips(
                vm_data["id"], nic_id
            ).get_result()
            floating_ips = res.get("floating_ips", [])
            for ip in floating_ips:
                if ip["name"].startswith(RAY_RECYCLABLE):
                    logger.debug(f"Deleting IP: {ip['id']}")
                    vpc_client.delete_floating_ip(ip["id"])
            logger.debug(f"Deleting VM: {vm_data['id']}")
            vpc_client.delete_instance(id=vm_data["id"])

        # pylint: disable=line-too-long E1136
        res = vpc_client.list_instances(vpc_id=vpc_id).get_result()
        num_instances = res["total_count"]

        # Delete VSIs if exist
        if num_instances:
            instances = res["instances"]
            with ThreadPoolExecutor(num_instances) as ex:
                for i in range(num_instances):
                    ex.submit(_del_instance, instances[i])
            # wait until all vms are deleted to proceed
            _poll_vpc_contains_vms(vpc_id)

    def delete_subnets(self, vpc_client, vpc_data, region):
        def _poll_subnet_deleted(subnet_id):
            tries = 10
            sleep_interval = 2
            while tries:
                try:
                    vpc_client.get_subnet(subnet_id).get_result()
                except ibm.ibm_cloud_sdk_core.ApiException:
                    logger.debug(f"Deleted subnet id: {subnet_id}")
                    return True
                tries -= 1
                time.sleep(sleep_interval)
            logger.error("Failed to delete instance within the allotted time\n")
            return False

        for subnet_id in self.get_vpc_subnets(vpc_data, region, field="id"):
            # get_result() used for synchronization
            logger.debug(f"Deleting subnet: {subnet_id}")
            vpc_client.delete_subnet(subnet_id).get_result()
            _poll_subnet_deleted(subnet_id)

    def delete_gateways(self, vpc_client, vpc_id):
        """deletes all gateways attached to the specified vpc"""
        # pylint: disable=line-too-long
        gateways = vpc_client.list_public_gateways(
            resource_group_id=self.resource_group_id
        ).get_result()["public_gateways"]
        gateways_ids_of_vpc = [
            gateway["id"] for gateway in gateways if gateway["vpc"]["id"] == vpc_id
        ]
        for gateway_id in gateways_ids_of_vpc:
            # get_result() used for synchronization
            logger.debug(f"Deleting gateway: {gateway_id}")
            vpc_client.delete_public_gateway(gateway_id).get_result()

    def add_missing_sg_rules(self, sec_group_id):
        missing_rules = self.get_unsatisfied_security_group_rules(sec_group_id)

        if missing_rules:
            for val in missing_rules.values():
                logger.debug(f"missing {val}")
            for missing_rule in missing_rules.keys():
                sg_rule_prototype = _build_security_group_rule_prototype_model(
                    missing_rule, sec_group_id
                )
                self.vpc_client.create_security_group_rule(
                    sec_group_id, sg_rule_prototype
                ).get_result()

    def get_unsatisfied_security_group_rules(self, sg_id):
        """
        returns unsatisfied security group rules.
        """

        unsatisfied_rules = copy.deepcopy(REQUIRED_RULES)
        sg = self.vpc_client.get_security_group(sg_id).get_result()

        for rule in sg["rules"]:
            # pylint: disable=line-too-long
            # check outbound rules that are not associated with a specific IP address range
            if rule["direction"] == "outbound" and rule["remote"] == {
                "cidr_block": "0.0.0.0/0"
            }:
                if rule["protocol"] == "all":
                    # outbound is fine!
                    unsatisfied_rules.pop("outbound_tcp_all", None)
                    unsatisfied_rules.pop("outbound_udp_all", None)
                elif rule["protocol"] == "tcp":
                    unsatisfied_rules.pop("outbound_tcp_all", None)
                elif rule["protocol"] == "udp":
                    unsatisfied_rules.pop("outbound_udp_all", None)

            # Check inbound rules
            elif rule["direction"] == "inbound":
                # check rules that are not associated with a specific IP address range
                if rule["remote"] == {"cidr_block": "0.0.0.0/0"}:
                    # we interested only in all or tcp protocols
                    if rule["protocol"] == "all":
                        # there a rule permitting all traffic
                        unsatisfied_rules.pop("inbound_tcp_sg", None)
                        unsatisfied_rules.pop("inbound_tcp_22", None)
                        unsatisfied_rules.pop("inbound_tcp_6379", None)
                        unsatisfied_rules.pop("inbound_tcp_8265", None)

                    elif rule["protocol"] == "tcp":
                        if rule["port_min"] == 1 and rule["port_max"] == 65535:
                            # all ports are open
                            unsatisfied_rules.pop("inbound_tcp_sg", None)
                            unsatisfied_rules.pop("inbound_tcp_22", None)
                            unsatisfied_rules.pop("inbound_tcp_6379", None)
                            unsatisfied_rules.pop("inbound_tcp_8265", None)
                        else:
                            port_min = rule["port_min"]
                            port_max = rule["port_max"]
                            if port_min <= 22 and port_max >= 22:
                                unsatisfied_rules.pop("inbound_tcp_22", None)
                            elif port_min <= 6379 and port_max >= 6379:
                                unsatisfied_rules.pop("inbound_tcp_6379", None)
                            elif port_min <= 8265 and port_max >= 8265:
                                unsatisfied_rules.pop("inbound_tcp_8265", None)

                # rule regards private traffic within the VSIs associated with the security group
                elif rule["remote"].get("id") == sg["id"]:
                    # validate that inbound traffic inside group available
                    if rule["protocol"] == "all" or rule["protocol"] == "tcp":
                        unsatisfied_rules.pop("inbound_tcp_sg", None)

        return unsatisfied_rules

    def remote_cluster_removal(self, vpc_id, region):
        """deletes the vpc and its associated resources
        using the FaaS service: ibm cloud functions.
        Used only when cluster is set to be deleted
        remotely, e.g. by `sky autostop --down`"""
        cc = ClusterCleaner(self.resource_group_id, vpc_id, region)
        cc.delete_cluster()


def _build_security_group_rule_prototype_model(missing_rule, sg_id=None):
    direction, protocol, port = missing_rule.split("_")
    remote = {"cidr_block": "0.0.0.0/0"}

    try:  # port number was specified
        port = int(port)
        port_min = port
        port_max = port
    # pylint: disable=W0703
    except Exception:
        port_min = 1
        port_max = 65535

        # only valid if security group already exists
        if port == "sg":
            if not sg_id:
                return None
            remote = {"id": sg_id}

    return {
        "direction": direction,
        "ip_version": "ipv4",
        "protocol": protocol,
        "remote": remote,
        "port_min": port_min,
        "port_max": port_max,
    }


class ClusterCleaner:
    """Responsible for deleting a cluster with all its associated
    resources. Used when the remote cluster head is requested to
    dismantle its cluster."""

    # default region for the cloud function namespace
    namespace_region = "us-east"
    # default name for the cloud function namespace (cf doesn't support tagging)
    namespace_name = "skypilot-namespace"
    # default name for the cloud function action.
    action_name = "skypilot-vpc-cleaner-action"
    # url to cloud function's namespaces in the chosen region: `namespace_region`
    cf_namespaces_url = (
        f"https://{namespace_region}.functions.cloud.ibm.com/api/v1/namespaces"
    )

    def __init__(self, resource_group_id, vpc_id, vpc_region) -> None:
        self.resource_group_id = resource_group_id
        self.vpc_id = vpc_id
        self.vpc_region = vpc_region

    function_code = textwrap.dedent(
        """
    import subprocess
    import time
    from concurrent.futures import ThreadPoolExecutor
    RAY_RECYCLABLE = "ray-recyclable"
    ibm_vpc_client = None
    # modules installed and imported entry point
    ibm_vpc = None
    ibm_cloud_sdk_core = None

    def get_vpc_data(vpc_id):

        if not vpc_id: return None
        try:
            vpc_data = ibm_vpc_client.get_vpc(vpc_id).result
            return vpc_data
        except ibm_cloud_sdk_core.ApiException as e:
            if e.code == 404:
                print(("VPC doesn't exist."))
                return None
            else: raise

    def delete_subnets(vpc_data):
        def _poll_subnet_exists(subnet_id):
            tries = 30 # waits up to 5 min with 10 sec interval
            sleep_interval = 10
            while tries:
                try:
                    subnet_data = ibm_vpc_client.get_subnet(subnet_id).result
                except Exception:
                    print('Deleted subnet id: {}'.format(subnet_id))
                    return True
                tries -= 1
                time.sleep(sleep_interval)
            print(f"Failed to delete instance within expected time frame of {tries*sleep_interval/60} minutes.")
            return False

        subnets_attached_to_routing_table = ibm_vpc_client.list_subnets(routing_table_id = vpc_data['default_routing_table']['id']).get_result()['subnets']
        subnets_ids = [subnet['id'] for subnet in subnets_attached_to_routing_table]
        for id in subnets_ids:
            try:
                ibm_vpc_client.delete_subnet(id).get_result()
                _poll_subnet_exists(id)
            except ibm_cloud_sdk_core.ApiException as e:
                if e.code == 404:
                    print("subnet doesn't exist.")

    def delete_gateways(vpc_id):
        gateways = ibm_vpc_client.list_public_gateways(resource_group_id=RESOURCE_GROUP_ID).get_result()['public_gateways']
        gateways_ids_of_vpc = [gateway['id'] for gateway in gateways if gateway['vpc']['id']== vpc_id]
        for gateway_id in gateways_ids_of_vpc:
            deleting_resource = True
            while deleting_resource:
                try:
                    ibm_vpc_client.delete_public_gateway(gateway_id).get_result()
                    deleting_resource = False
                except ibm_cloud_sdk_core.ApiException as e:
                    if e.code == 404:
                        print("gateway doesn't exist.")
                        deleting_resource = False
                    if e.code == 409:
                        print("gateway still in use.")
                        # will retry until cloud functions timeout.
                        time.sleep(5)

    def delete_vms(vpc_id):
        def _poll_vpc_contains_vms(vpc_id):
            tries = 60
            sleep_interval = 3
            while tries:
                # list_instances() never raise an exception, check values instead
                res = ibm_vpc_client.list_instances(vpc_id=vpc_id).get_result()
                if not res["total_count"]:
                    return True
                else:
                    tries -= 1
                    time.sleep(sleep_interval)
            raise Exception(
                "Failed to delete VPC's instances within "
                "the expected time frame. Cannot "
                "continue to delete VPC."
            )

        def _del_instance(vm_data):
            # first delete ips created by node_provider
            nic_id = vm_data["network_interfaces"][0]["id"]
            res = ibm_vpc_client.list_instance_network_interface_floating_ips(
                vm_data["id"], nic_id
            ).get_result()
            floating_ips = res.get("floating_ips", [])
            for ip in floating_ips:
                if ip["name"].startswith(RAY_RECYCLABLE):
                    print(f"Deleting IP: {ip['id']}")
                    ibm_vpc_client.delete_floating_ip(ip["id"])
            print(f"Deleting VM: {vm_data['id']}")
            ibm_vpc_client.delete_instance(id=vm_data["id"])

        res = ibm_vpc_client.list_instances(vpc_id=vpc_id).get_result()
        num_instances = res["total_count"]

        # Delete VSIs if exist
        if num_instances:
            instances = res["instances"]
            with ThreadPoolExecutor(num_instances) as ex:
                for i in range(num_instances):
                    ex.submit(_del_instance, instances[i])
            # wait until all vms are deleted to proceed
            _poll_vpc_contains_vms(vpc_id)

    def delete_unbound_vpc(vpc_id):
        deleting_resource = True
        while deleting_resource:
            try:
                ibm_vpc_client.delete_vpc(vpc_id).get_result()
                deleting_resource = False
            except ibm_cloud_sdk_core.ApiException as e:
                if e.code == 404:
                    print("VPC doesn't exist.")
                    deleting_resource = False
                if e.code == 409:
                    print("VPC still in use.")
                    # will retry until cloud functions timeout.
                    time.sleep(5)

    def delete_vpc(vpc_id):
        vpc_data = get_vpc_data(vpc_id)
        if not vpc_data:
            print((f"Failed to find a VPC with id={vpc_id}"))
            return
        print(f"Deleting vpc:{vpc_data['name']} with id:{vpc_id}")
        delete_vms(vpc_id)
        delete_subnets(vpc_data)
        delete_gateways(vpc_id)
        delete_unbound_vpc(vpc_id)
        print(f"VPC {vpc_data['name']} and its attached resources were deleted successfully")


    def main(dict):
        global ibm_vpc_client, RESOURCE_GROUP_ID, ibm_cloud_sdk_core, ibm_vpc
        def install_package(package):
            pip_location_stdout = subprocess.run(['which', 'pip'], capture_output=True, text=True)
            pip_location = pip_location_stdout.stdout.strip()
            subprocess.call([pip_location, 'install', package], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)

        for package in ["ibm-vpc", "ibm-cloud-sdk-core"]:
            install_package(package)

        import ibm_vpc as _ibm_vpc
        import ibm_cloud_sdk_core as _ibm_cloud_sdk_core
        ibm_vpc = _ibm_vpc
        ibm_cloud_sdk_core = _ibm_cloud_sdk_core

        iam_api_key, RESOURCE_GROUP_ID, vpc_id, region = dict['iam_api_key'], dict['resource_group_id'], dict['vpc_id'], dict['region']

        authenticator = ibm_cloud_sdk_core.authenticators.IAMAuthenticator(iam_api_key, url=None)
        ibm_vpc_client = ibm_vpc.VpcV1('2022-06-30',authenticator=authenticator)

        if not region:
            raise Exception("VPC not found in any region")

        ibm_vpc_client.set_service_url(f'https://{region}.iaas.cloud.ibm.com/v1')

        delete_vpc(vpc_id=vpc_id)
        return {"Status": "Success"}
    """
    )

    def get_headers(self):
        return {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer " + ibm.get_oauth_token(),
        }

    def create_or_fetch_namespace(self):
        """returns namespace id.
        creates a namespace with given name in specified region.
        if namespace exists returns its id instead."""

        def _create_new_namespace():
            logger.info(
                f"Creating a new namespace: {self.namespace_name} in {self.namespace_region}"
            )

            data = {
                "name": self.namespace_name,
                "resource_group_id": self.resource_group_id,
                "resource_plan_id": "functions-base-plan",
            }

            res = requests.post(
                self.cf_namespaces_url, headers=self.get_headers(), json=data
            ).json()
            if res.status_code != 200:
                logger.error(res.text)
            namespace_id = res["id"]
            logger.info(f"Created new namespace with id: {namespace_id}")
            return namespace_id

        def _get_cloud_function_namespaces_metadata(offset=0):
            """returns meta data on namespaces of ibm cloud functions within a specified region
            :param offset - offset from the beginning of the list of results attained from the GET request,
                            which may contain up to 200 namespaces per http response"""

            res = requests.get(
                f"{self.cf_namespaces_url}?limit=200&offset={offset}",
                headers=self.get_headers(),
            )
            return json.loads(res.text)

        def _get_cloud_function_namespaces():
            """returns relevant metadata on existing namespaces within a given region."""
            logger.info(
                f"Obtaining Cloud Function namespaces in {self.namespace_region}"
            )

            namespaces = []

            collecting_namespaces = True
            max_limit = 200
            offset = 0

            #  request for namespaces is limited to 200 at a time, thus the request is fulfilled in increments of 200s.
            while collecting_namespaces:
                namespace_metadata = _get_cloud_function_namespaces_metadata(offset)
                if namespace_metadata["total_count"] == max_limit:
                    offset += max_limit
                else:
                    collecting_namespaces = False

                for name_space in namespace_metadata["namespaces"]:
                    if "name" in name_space:  # API based namespace
                        namespaces.append(
                            {
                                "name": name_space["name"],
                                "type": "API_based",
                                "id": name_space["id"],
                                "region": name_space["location"],
                            }
                        )

                    else:  # cloud foundry based namespace
                        namespaces.append(
                            {
                                "name": name_space["id"],
                                "type": "CF_based",
                                "region": name_space["location"],
                            }
                        )

            return namespaces

        namespaces_in_region = _get_cloud_function_namespaces()
        target_namespace_id = None
        if namespaces_in_region:
            target_namespace_id = next(
                (
                    namespace["id"]
                    for namespace in namespaces_in_region
                    if namespace["name"] == self.namespace_name
                ),
                None,
            )
        if not target_namespace_id:
            target_namespace_id = _create_new_namespace()
        else:
            logger.info(f"Reusing namespace: {target_namespace_id}")
        return target_namespace_id

    def _get_cloud_functions_actions(self, namespace_id):
        """returns meta data on namespaces of ibm cloud functions within a specified region
        :param offset - offset from the beginning of the list of results attained from the GET request,
                        which may contain up to 200 namespaces per http response"""

        res = requests.get(
            f"{self.cf_namespaces_url}/{namespace_id}/actions?limit=200",
            headers=self.get_headers(),
        )
        return json.loads(res.text)

    def create_action(self, namespace_id):
        logger.info(f"creating action on namespace: {namespace_id}")
        # Define the function parameters
        function_params = {
            "exec": {"kind": "python:3.9", "code": self.function_code},
            "limits": {"timeout": 600000},
        }
        res = requests.put(
            f"{self.cf_namespaces_url}/{namespace_id}/actions/{self.action_name}?blocking=true&overwrite=true",
            headers=self.get_headers(),
            data=json.dumps(function_params),
        )
        if res.status_code != 200:
            logger.error(res.text)
        return json.loads(res.text)

    def delete_action(self, namespace_id):
        """return the deleted function's metadata if it existed."""
        logger.info(f"deleting action on namespace: {namespace_id}")
        res = requests.delete(
            f"{self.cf_namespaces_url}/{namespace_id}/actions/{self.action_name}?blocking=true",
            headers=self.get_headers(),
        )
        if res.status_code != 200:
            logger.warn(res.text)
        return json.loads(res.text)

    def invoke_action(self, namespace_id: str):
        logger.info(f"invoking action on namespace: {namespace_id}")
        payload = {
            "iam_api_key": ibm.get_api_key(),
            "resource_group_id": self.resource_group_id,
            "vpc_id": self.vpc_id,
            "region": self.vpc_region,
        }
        res = requests.post(
            f"{self.cf_namespaces_url}/{namespace_id}/actions/{self.action_name}?blocking=true",
            headers=self.get_headers(),
            data=json.dumps(payload),
        )
        if res.status_code != 200:
            logger.error(res.text)
        return json.loads(res.text)

    def delete_cluster(self):
        """Deletes VPC with id==self.vpc_id.
        1. creates a CloudFunctions namespace named ClusterCleaner.namespace_name if doesn't exists.
        2. using idempotent function that deletes an action named ClusterCleaner.action_name if exists.
        3. invokes the action to delete the VPC and all its resources."""
        cf_namespace_id = self.create_or_fetch_namespace()
        self.delete_action(cf_namespace_id)
        self.create_action(cf_namespace_id)
        self.invoke_action(cf_namespace_id)



"""Shadeform cloud adaptor."""

import functools
import socket
from typing import Any, Dict, List, Optional

import requests

from sky import sky_logging
from sky.provision.shadeform import shadeform_utils
from sky.utils import common_utils

logger = sky_logging.init_logger(__name__)

_shadeform_sdk = None


def import_package(func):

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        global _shadeform_sdk
        if _shadeform_sdk is None:
            try:
                import shadeform as _shadeform  # pylint: disable=import-outside-toplevel
                _shadeform_sdk = _shadeform
            except ImportError:
                raise ImportError(
                    'Failed to import dependencies for Shadeform. '
                    'Try pip install "skypilot[shadeform]"') from None
        return func(*args, **kwargs)

    return wrapper


@import_package
def shadeform():
    """Return the shadeform package."""
    return _shadeform_sdk


def list_ssh_keys() -> List[Dict[str, Any]]:
    """List all SSH keys in Shadeform account."""
    try:
        response = shadeform_utils.get_ssh_keys()
        return response.get('ssh_keys', [])
    except (ValueError, KeyError, requests.exceptions.RequestException) as e:
        logger.warning(f'Failed to list SSH keys from Shadeform: {e}')
        return []


def add_ssh_key_to_shadeform(public_key: str) -> Optional[str]:
    """Add SSH key to Shadeform if it doesn't already exist.

    Args:
        public_key: The SSH public key string.

    Returns:
        The name of the key if added successfully, None otherwise.
    """
    try:
        # Check if key already exists
        existing_keys = list_ssh_keys()
        key_exists = False
        key_id = None
        for key in existing_keys:
            if key.get('public_key') == public_key:
                key_exists = True
                key_id = key.get('id')
                break

        if key_exists:
            logger.info('SSH key already exists in Shadeform account')
            return key_id

        # Generate a unique key name
        hostname = socket.gethostname()
        key_name = f'skypilot-{hostname}-{common_utils.get_user_hash()[:8]}'

        # Add the key
        response = shadeform_utils.add_ssh_key(name=key_name,
                                               public_key=public_key)
        key_id = response['id']
        logger.info(f'Added SSH key to Shadeform: {key_name, key_id}')
        return key_id

    except (ValueError, KeyError, requests.exceptions.RequestException) as e:
        logger.warning(f'Failed to add SSH key to Shadeform: {e}')
        return None



"""Code generator for task execution."""

import copy
import inspect
import json
import math
import os
import shlex
import textwrap
from typing import Dict, List, Optional, Tuple

import colorama

from sky import sky_logging
from sky.skylet import constants
from sky.skylet import log_lib
from sky.utils import accelerator_registry
from sky.utils import ux_utils

# Unset RAY_RAYLET_PID to prevent the Ray cluster in the SkyPilot runtime
# from interfering with the Ray cluster in the user's task (if any).
UNSET_RAY_ENV_VARS = ['RAY_RAYLET_PID']

logger = sky_logging.init_logger(__name__)


class TaskCodeGen:
    """Base code generator for task execution on Ray and Slurm."""

    def __init__(self) -> None:
        # Code generated so far, to be joined via '\n'.
        self._code: List[str] = []
        # Guard method calling order.
        self._has_prologue: bool = False
        self._has_epilogue: bool = False
        self._has_setup: bool = False
        # Job ID is used to identify the job (also this generated code).
        self.job_id: Optional[int] = None

    def _add_common_imports(self) -> None:
        """Add common imports for both Ray and Slurm execution."""
        self._code.append(
            textwrap.dedent("""\
            import functools
            import getpass
            import hashlib
            import io
            import os
            import pathlib
            import selectors
            import shlex
            import subprocess
            import sys
            import tempfile
            import textwrap
            import time
            from typing import Dict, List, Optional, Tuple, Union
            """))

    def _add_skylet_imports(self) -> None:
        """Add SkyPilot skylet imports."""
        self._code.append(
            textwrap.dedent("""\
            from sky.skylet import autostop_lib
            from sky.skylet import constants
            from sky.skylet import job_lib
            from sky.utils import log_utils
            from sky.utils import subprocess_utils
            """))

    def _add_logging_functions(self) -> None:
        """Add log streaming functions from log_lib."""
        self._code += [
            # FIXME: This is a hack to make sure that the functions can be found
            # by ray.remote. This should be removed once we have a better way to
            # specify dependencies for ray.
            inspect.getsource(log_lib._ProcessingArgs),  # pylint: disable=protected-access
            inspect.getsource(log_lib._get_context),  # pylint: disable=protected-access
            inspect.getsource(log_lib._handle_io_stream),  # pylint: disable=protected-access
            inspect.getsource(log_lib.process_subprocess_stream),
            inspect.getsource(log_lib.run_with_log),
            inspect.getsource(log_lib.make_task_bash_script),
            inspect.getsource(log_lib.add_ray_env_vars),
            inspect.getsource(log_lib.run_bash_command_with_log),
            inspect.getsource(log_lib.run_bash_command_with_log_and_return_pid),
        ]

    def _add_waiting_for_resources_msg(self, num_nodes: int) -> None:
        self._code.append(
            textwrap.dedent(f"""\
            plural = 's' if {num_nodes} > 1 else ''
            node_str = f'{num_nodes} node{{plural}}'
            message = ('{ux_utils.INDENT_SYMBOL}{colorama.Style.DIM}'
                       'Waiting for task resources on '
                       f'{{node_str}}.{colorama.Style.RESET_ALL}')
            print(message, flush=True)"""))

    def _get_job_started_msg(self) -> str:
        """Returns the 'Job started' streaming message with ANSI formatting."""
        return (
            f'{ux_utils.INDENT_LAST_SYMBOL}Job started. Streaming logs... '
            f'{colorama.Style.DIM}(Ctrl-C to exit log streaming; job will not '
            f'be killed){colorama.Style.RESET_ALL}')

    def _add_job_started_msg(self) -> None:
        streaming_message = self._get_job_started_msg()
        self._code.append(f'print({streaming_message!r}, flush=True)')

    def _get_accelerator_details(
        self,
        resources_dict: Dict[str, float],
    ) -> Tuple[Optional[str], float]:
        resources_copy = resources_dict.copy()
        resources_copy.pop('CPU', None)

        if not resources_copy:
            return None, 0.0

        assert len(resources_copy) == 1, (
            'There can only be one type of accelerator per instance. '
            f'Found: {resources_copy}.')

        acc_name, acc_count = list(resources_copy.items())[0]
        return acc_name, float(acc_count)

    def _add_constants(self) -> None:
        self._code.append(
            textwrap.dedent(f"""\
            SKY_REMOTE_WORKDIR = {constants.SKY_REMOTE_WORKDIR!r}

            CANCELLED_RETURN_CODE = 137
            """))

    @staticmethod
    def get_rclone_flush_script() -> str:
        """Generate rclone flush script for cached storage mounts.

        This script blocks job completion until all storage mounted with
        CACHED_MOUNT mode is uploaded to remote.

        Returns:
            Bash script as string
        """
        return textwrap.dedent(f"""\

        # Only waits if cached mount is enabled (RCLONE_MOUNT_CACHED_LOG_DIR is not empty)
        # findmnt alone is not enough, as some clouds (e.g. AWS on ARM64) uses
        # rclone for normal mounts as well.
        if [ $(findmnt -t fuse.rclone --noheading | wc -l) -gt 0 ] && \
           [ -d {constants.RCLONE_MOUNT_CACHED_LOG_DIR} ] && \
           [ "$(ls -A {constants.RCLONE_MOUNT_CACHED_LOG_DIR})" ]; then
            FLUSH_START_TIME=$(date +%s)
            flushed=0
            # extra second on top of --vfs-cache-poll-interval to
            # avoid race condition between rclone log line creation and this check.
            sleep 1
            while [ $flushed -eq 0 ]; do
                # sleep for the same interval as --vfs-cache-poll-interval
                sleep {constants.RCLONE_CACHE_REFRESH_INTERVAL}
                flushed=1
                for file in {constants.RCLONE_MOUNT_CACHED_LOG_DIR}/*; do
                    exitcode=0
                    tac $file | grep "vfs cache: cleaned:" -m 1 | grep "in use 0, to upload 0, uploading 0" -q || exitcode=$?
                    if [ $exitcode -ne 0 ]; then
                        ELAPSED=$(($(date +%s) - FLUSH_START_TIME))
                        # Extract the last vfs cache status line to show what we're waiting for
                        CACHE_STATUS=$(tac $file | grep "vfs cache: cleaned:" -m 1 | sed 's/.*vfs cache: cleaned: //' 2>/dev/null)
                        # Extract currently uploading files from recent log lines (show up to 2 files)
                        UPLOADING_FILES=$(tac $file | head -30 | grep -E "queuing for upload" | head -2 | sed 's/.*INFO  : //' | sed 's/: vfs cache:.*//' | tr '\\n' ',' | sed 's/,$//' | sed 's/,/, /g' 2>/dev/null)
                        # Build status message with available info
                        if [ -n "$CACHE_STATUS" ] && [ -n "$UPLOADING_FILES" ]; then
                            echo "skypilot: cached mount is still uploading (elapsed: ${{ELAPSED}}s) [${{CACHE_STATUS}}] uploading: ${{UPLOADING_FILES}}"
                        elif [ -n "$CACHE_STATUS" ]; then
                            echo "skypilot: cached mount is still uploading (elapsed: ${{ELAPSED}}s) [${{CACHE_STATUS}}]"
                        else
                            # Fallback: show last non-empty line from log
                            LAST_LINE=$(tac $file | grep -v "^$" | head -1 | sed 's/.*INFO  : //' | sed 's/.*ERROR : //' | sed 's/.*NOTICE: //' 2>/dev/null)
                            if [ -n "$LAST_LINE" ]; then
                                echo "skypilot: cached mount is still uploading (elapsed: ${{ELAPSED}}s) ${{LAST_LINE}}"
                            else
                                echo "skypilot: cached mount is still uploading (elapsed: ${{ELAPSED}}s)"
                            fi
                        fi
                        flushed=0
                        break
                    fi
                done
            done
            TOTAL_FLUSH_TIME=$(($(date +%s) - FLUSH_START_TIME))
            echo "skypilot: cached mount upload complete (took ${{TOTAL_FLUSH_TIME}}s)"
        fi""")

    def add_prologue(self, job_id: int) -> None:
        """Initialize code generator and add prologue code.

        Args:
            job_id: SkyPilot internal job ID
        """
        raise NotImplementedError

    def add_setup(
        self,
        num_nodes: int,
        resources_dict: Dict[str, float],
        stable_cluster_internal_ips: List[str],
        env_vars: Dict[str, str],
        log_dir: str,
        setup_cmd: Optional[str] = None,
    ) -> None:
        """Generates code to set up the task on each node.

        stable_cluster_internal_ips is used to ensure that the
        SKYPILOT_NODE_RANK environment variable is assigned in a
        deterministic order whenever a new task is added.
        """
        raise NotImplementedError

    def add_task(
        self,
        num_nodes: int,
        bash_script: Optional[str],
        task_name: Optional[str],
        resources_dict: Dict[str, float],
        log_dir: str,
        env_vars: Optional[Dict[str, str]] = None,
    ) -> None:
        """Generates code to run the bash command on all num_nodes nodes."""
        raise NotImplementedError

    def add_epilogue(self) -> None:
        """Generate code that checks return codes and updates job status."""
        assert self._has_prologue, 'Call add_prologue() before add_epilogue().'
        assert not self._has_epilogue, 'add_epilogue() called twice?'
        self._has_epilogue = True

        self._code += [
            textwrap.dedent(f"""\
            if sum(returncodes) != 0:
                # Save exit codes to job metadata for potential recovery logic
                if int(constants.SKYLET_VERSION) >= 28:
                    job_lib.set_exit_codes({self.job_id!r}, returncodes)
                job_lib.set_status({self.job_id!r}, job_lib.JobStatus.FAILED)
                # Schedule the next pending job immediately to make the job
                # scheduling more efficient.
                job_lib.scheduler.schedule_step()
                # This waits for all streaming logs to finish.
                time.sleep(0.5)
                reason = ''
                # 139 is the return code of SIGSEGV, i.e. Segmentation Fault.
                if any(r == 139 for r in returncodes):
                    reason = '(likely due to Segmentation Fault)'
                if any(r == 137 for r in returncodes):
                    # Find the first non-137 return code
                    non_137 = next(r for r in returncodes if r != 137)
                    reason = f'(A Worker failed with return code {{non_137}}, SkyPilot cleaned up the processes on other nodes with return code 137)'
                print('ERROR: {colorama.Fore.RED}Job {self.job_id} failed with '
                      'return code list:{colorama.Style.RESET_ALL}',
                      returncodes,
                      reason,
                      flush=True)
                # Need this to set the job status in ray job to be FAILED.
                sys.exit(1)
            else:
                job_lib.set_status({self.job_id!r}, job_lib.JobStatus.SUCCEEDED)
                # Schedule the next pending job immediately to make the job
                # scheduling more efficient.
                job_lib.scheduler.schedule_step()
                # This waits for all streaming logs to finish.
                time.sleep(0.5)
            """)
        ]

    def build(self) -> str:
        """Returns the entire generated program."""
        assert self._has_epilogue, 'Call add_epilogue() before build().'
        return '\n'.join(self._code)


class RayCodeGen(TaskCodeGen):
    """Code generator of a Ray program that executes a sky.Task.

    Usage:

      >> codegen = RayCodegen()
      >> codegen.add_prologue()

      >> codegen.add_task(...)
      >> codegen.add_task(...)

      >> codegen.add_epilogue()
      >> code = codegen.build()
    """

    def add_prologue(self, job_id: int) -> None:
        assert not self._has_prologue, 'add_prologue() called twice?'
        self._has_prologue = True
        self.job_id = job_id
        # Should use 'auto' or 'ray://<internal_head_ip>:10001' rather than
        # 'ray://localhost:10001', or 'ray://127.0.0.1:10001', for public cloud.
        # Otherwise, ray will fail to get the placement group because of a bug
        # in ray job.
        ray_address = 'auto'

        # Add common imports
        self._add_common_imports()

        # Add Ray-specific setup
        self._code.append(
            textwrap.dedent("""\
            # Set the environment variables to avoid deduplicating logs and
            # scheduler events. This should be set in driver code, since we are
            # not using `ray job submit` anymore, and the environment variables
            # from the ray cluster is not inherited.
            os.environ['RAY_DEDUP_LOGS'] = '0'
            os.environ['RAY_SCHEDULER_EVENTS'] = '0'

            import ray
            import ray.util as ray_util
            """))

        self._add_skylet_imports()

        self._add_constants()

        # Add Ray configuration
        self._code.append(
            textwrap.dedent(f"""\
            kwargs = dict()
            # Only set the `_temp_dir` to SkyPilot's ray cluster directory when
            # the directory exists for backward compatibility for the VM
            # launched before #1790.
            if os.path.exists({constants.SKY_REMOTE_RAY_TEMPDIR!r}):
                kwargs['_temp_dir'] = {constants.SKY_REMOTE_RAY_TEMPDIR!r}
            ray.init(
                address={ray_address!r},
                namespace='__sky__{job_id}__',
                log_to_driver=True,
                **kwargs
            )
            def get_or_fail(futures, pg) -> List[int]:
                \"\"\"Wait for tasks, if any fails, cancel all unready.\"\"\"
                if not futures:
                    return [], []
                returncodes = [1] * len(futures)
                pids = [None] * len(futures)
                failed = False
                # Wait for 1 task to be ready.
                ready = []
                # Keep invoking ray.wait if ready is empty. This is because
                # ray.wait with timeout=None will only wait for 10**6 seconds,
                # which will cause tasks running for more than 12 days to return
                # before becoming ready.
                # (Such tasks are common in serving jobs.)
                # Reference: https://github.com/ray-project/ray/blob/ray-2.9.3/python/ray/_private/worker.py#L2845-L2846

                def handle_ready_tasks(tasks: List[ray.ObjectRef]) -> None:
                    nonlocal returncodes, pids, failed
                    for task in tasks:
                        idx = futures.index(task)
                        res = ray.get(task)
                        returncodes[idx] = res['return_code']
                        pids[idx] = res['pid']
                        if res['return_code'] != 0:
                            failed = True

                while not ready:
                    ready, unready = ray.wait(futures)
                handle_ready_tasks(ready)
                while unready:
                    if failed:
                        for task in unready:
                            # ray.cancel without force fails to kill tasks.
                            # We use force=True to kill unready tasks.
                            ray.cancel(task, force=True)
                            # Use SIGKILL=128+9 to indicate the task is forcely
                            # killed.
                            idx = futures.index(task)
                            returncodes[idx] = CANCELLED_RETURN_CODE
                        break
                    ready, unready = ray.wait(unready)
                    handle_ready_tasks(ready)
                # Remove the placement group after all tasks are done, so that
                # the next job can be scheduled on the released resources
                # immediately.
                ray_util.remove_placement_group(pg)
                sys.stdout.flush()
                return returncodes, pids

            futures = []
            """))

        self._add_logging_functions()

        self._code += [
            'run_bash_command_with_log = run_bash_command_with_log',
            'run_bash_command_with_log_and_return_pid = \
                ray.remote(run_bash_command_with_log_and_return_pid)',
            'autostop_lib.set_last_active_time_to_now()',
            f'job_lib.set_status({job_id!r}, job_lib.JobStatus.PENDING)',
        ]

    def add_setup(
        self,
        num_nodes: int,
        resources_dict: Dict[str, float],
        stable_cluster_internal_ips: List[str],
        env_vars: Dict[str, str],
        log_dir: str,
        setup_cmd: Optional[str] = None,
    ) -> None:
        assert self._has_prologue, ('Call add_prologue() before '
                                    'add_setup().')
        self._has_setup = True

        setup_log_path = os.path.join(log_dir, 'setup.log')

        bundles = [copy.copy(resources_dict) for _ in range(num_nodes)]
        # Set CPU to avoid ray hanging the resources allocation
        # for remote functions, since the task will request 1 CPU
        # by default.
        task_cpu_demand = resources_dict.pop('CPU')

        if resources_dict:
            assert len(resources_dict) == 1, (
                'There can only be one type of accelerator per instance. '
                f'Found: {resources_dict}.')
            acc_name, acc_count = list(resources_dict.items())[0]
            gpu_dict = {'GPU': acc_count}
            # gpu_dict should be empty when the accelerator is not GPU.
            # TODO(zongheng,zhanghao): an alternative is to start the remote
            # cluster with custom resource 'GPU': <n> even if the accelerator(s)
            # are not GPU. We opt for the current solution for now.
            if accelerator_registry.is_schedulable_non_gpu_accelerator(
                    acc_name):
                gpu_dict = {}
            for bundle in bundles:
                bundle.update({
                    # Set the GPU to avoid ray hanging the resources allocation
                    **gpu_dict,
                })

        self._code.append(
            f'pg = ray_util.placement_group({json.dumps(bundles)}, '
            f'\'STRICT_SPREAD\')')
        self._add_waiting_for_resources_msg(num_nodes)
        self._code.append(
            textwrap.dedent("""\
            # FIXME: This will print the error message from autoscaler if
            # it is waiting for other task to finish. We should hide the
            # error message.
            ray.get(pg.ready())"""))
        self._add_job_started_msg()

        job_id = self.job_id
        if setup_cmd is not None:
            setup_envs = env_vars.copy()
            setup_envs[constants.SKYPILOT_NUM_NODES] = str(num_nodes)
            self._code += [
                textwrap.dedent(f"""\
                setup_cmd = {setup_cmd!r}
                _SETUP_CPUS = 0.0001
                # The setup command will be run as a ray task with num_cpus=_SETUP_CPUS as the
                # requirement; this means Ray will set CUDA_VISIBLE_DEVICES to an empty string.
                # We unset it so that user setup command may properly use this env var.
                setup_cmd = 'unset CUDA_VISIBLE_DEVICES; ' + setup_cmd
                job_lib.set_status({job_id!r}, job_lib.JobStatus.SETTING_UP)

                # The schedule_step should be called after the job status is set to non-PENDING,
                # otherwise, the scheduler will think the current job is not submitted yet, and
                # skip the scheduling step.
                job_lib.scheduler.schedule_step()

                # If some nodes are down and then new nodes are added after launching again,
                # the result of `ray.nodes()` will include all the nodes, so we need to get
                # the alive nodes.
                alive_nodes = [n for n in ray.nodes() if 'Alive' in n and n['Alive']]
                total_num_nodes = len(alive_nodes)
                setup_bundles = [{{"CPU": _SETUP_CPUS}} for _ in range(total_num_nodes)]
                setup_pg = ray.util.placement_group(setup_bundles, strategy='STRICT_SPREAD')
                setup_workers = [run_bash_command_with_log_and_return_pid \\
                    .options(
                        name='setup',
                        num_cpus=_SETUP_CPUS,
                        scheduling_strategy=ray.util.scheduling_strategies.PlacementGroupSchedulingStrategy(
                            placement_group=setup_pg,
                            placement_group_bundle_index=i)
                    ) \\
                    .remote(
                        setup_cmd,
                        os.path.expanduser({setup_log_path!r}),
                        env_vars={setup_envs!r},
                        stream_logs=True,
                        with_ray=True,
                    ) for i in range(total_num_nodes)]
                setup_returncodes, setup_pids = get_or_fail(setup_workers, setup_pg)
                success = True
                failed_workers_and_returncodes = []
                for i in range(len(setup_returncodes)):
                    returncode = setup_returncodes[i]
                    pid = setup_pids[i]
                    if pid == None:
                        pid = os.getpid()
                    if returncode != 0 and returncode != CANCELLED_RETURN_CODE:
                        success = False
                        failed_workers_and_returncodes.append((pid, returncode))
                if not success:
                    msg = f'ERROR: {colorama.Fore.RED}Job {self.job_id}\\'s setup failed. '
                    msg += f'Failed workers: ' + ', '.join([f'(pid={{pid}}, returncode={{returncode}})' for pid, returncode in failed_workers_and_returncodes])
                    msg += f'. See error logs above for more details.{colorama.Style.RESET_ALL}'
                    print(msg, flush=True)
                    if int(constants.SKYLET_VERSION) >= 28:
                        job_lib.set_exit_codes({self.job_id!r}, setup_returncodes)
                    job_lib.set_status({self.job_id!r}, job_lib.JobStatus.FAILED_SETUP)
                    # This waits for all streaming logs to finish.
                    time.sleep(1)
                    # Need this to set the job status in ray job to be FAILED.
                    sys.exit(1)
                """)
            ]

        self._code.append(f'job_lib.set_job_started({self.job_id!r})')
        if setup_cmd is None:
            # Need to call schedule_step() to make sure the scheduler
            # schedule the next pending job.
            self._code.append('job_lib.scheduler.schedule_step()')

        # Export IP and node rank to the environment variables.
        self._code += [
            textwrap.dedent(f"""\
                @ray.remote
                def check_ip():
                    return ray.util.get_node_ip_address()
                gang_scheduling_id_to_ip = ray.get([
                    check_ip.options(
                            num_cpus={task_cpu_demand},
                            scheduling_strategy=ray.util.scheduling_strategies.PlacementGroupSchedulingStrategy(
                                placement_group=pg,
                                placement_group_bundle_index=i
                            )).remote()
                    for i in range(pg.bundle_count)
                ])

                cluster_ips_to_node_id = {{ip: i for i, ip in enumerate({stable_cluster_internal_ips!r})}}
                job_ip_rank_list = sorted(gang_scheduling_id_to_ip, key=cluster_ips_to_node_id.get)
                job_ip_rank_map = {{ip: i for i, ip in enumerate(job_ip_rank_list)}}
                job_ip_list_str = '\\n'.join(job_ip_rank_list)
                """),
        ]

    def add_task(self,
                 num_nodes: int,
                 bash_script: Optional[str],
                 task_name: Optional[str],
                 resources_dict: Dict[str, float],
                 log_dir: str,
                 env_vars: Optional[Dict[str, str]] = None) -> None:
        # TODO(zhwu): The resources limitation for multi-node ray.tune and
        # horovod should be considered.
        for i in range(num_nodes):
            # Ray's per-node resources, to constrain scheduling each command to
            # the corresponding node, represented by private IPs.
            self._add_ray_task(bash_script=bash_script,
                               task_name=task_name,
                               resources_dict=resources_dict.copy(),
                               log_dir=log_dir,
                               env_vars=env_vars,
                               gang_scheduling_id=i)

    def _add_ray_task(self,
                      bash_script: Optional[str],
                      task_name: Optional[str],
                      resources_dict: Dict[str, float],
                      log_dir: str,
                      env_vars: Optional[Dict[str, str]] = None,
                      gang_scheduling_id: int = 0) -> None:
        """Generates code for a ray remote task that runs a bash command."""
        assert self._has_setup, 'Call add_setup() before add_task().'

        task_cpu_demand = resources_dict.pop('CPU')
        # Build remote_task.options(...)
        #   resources=...
        #   num_gpus=...
        options = []
        options.append(f'num_cpus={task_cpu_demand}')

        acc_name, acc_count = self._get_accelerator_details(resources_dict)
        num_gpus = 0.0
        if acc_name is not None:
            assert resources_dict, ('There can only be one type of accelerator '
                                    'per instance.')
            options.append(f'resources={json.dumps(resources_dict)}')
            if not accelerator_registry.is_schedulable_non_gpu_accelerator(
                    acc_name):
                num_gpus = acc_count
                options.append(f'num_gpus={num_gpus}')
        options.append(
            'scheduling_strategy=ray.util.scheduling_strategies.PlacementGroupSchedulingStrategy('  # pylint: disable=line-too-long
            'placement_group=pg, '
            f'placement_group_bundle_index={gang_scheduling_id})')

        sky_env_vars_dict_str = [
            textwrap.dedent(f"""\
            sky_env_vars_dict = {{}}
            sky_env_vars_dict['{constants.SKYPILOT_NODE_IPS}'] = job_ip_list_str
            sky_env_vars_dict['{constants.SKYPILOT_NUM_NODES}'] = len(job_ip_rank_list)
            """)
        ]

        if env_vars is not None:
            sky_env_vars_dict_str.extend(f'sky_env_vars_dict[{k!r}] = {v!r}'
                                         for k, v in env_vars.items())
        sky_env_vars_dict_str = '\n'.join(sky_env_vars_dict_str)

        options_str = ', '.join(options)
        logger.debug('Added Task with options: '
                     f'{options_str}')
        rclone_flush_script = self.get_rclone_flush_script()
        unset_ray_env_vars = ' && '.join(
            [f'unset {var}' for var in UNSET_RAY_ENV_VARS])
        self._code += [
            sky_env_vars_dict_str,
            textwrap.dedent(f"""\
        script = {bash_script!r}
        rclone_flush_script = {rclone_flush_script!r}

        if script is not None:
            script=f'{unset_ray_env_vars}; {{script}}'
            script += rclone_flush_script
            sky_env_vars_dict['{constants.SKYPILOT_NUM_GPUS_PER_NODE}'] = {int(math.ceil(num_gpus))!r}

            ip = gang_scheduling_id_to_ip[{gang_scheduling_id!r}]
            rank = job_ip_rank_map[ip]

            if len(cluster_ips_to_node_id) == 1: # Single-node task on single-node cluter
                name_str = '{task_name},' if {task_name!r} != None else 'task,'
                log_path = os.path.expanduser(os.path.join({log_dir!r}, 'run.log'))
            else: # Single-node or multi-node task on multi-node cluster
                idx_in_cluster = cluster_ips_to_node_id[ip]
                if cluster_ips_to_node_id[ip] == 0:
                    node_name = 'head'
                else:
                    node_name = f'worker{{idx_in_cluster}}'
                name_str = f'{{node_name}}, rank={{rank}},'
                log_path = os.path.expanduser(os.path.join({log_dir!r}, f'{{rank}}-{{node_name}}.log'))
            sky_env_vars_dict['{constants.SKYPILOT_NODE_RANK}'] = rank

            sky_env_vars_dict['SKYPILOT_INTERNAL_JOB_ID'] = {self.job_id}

            futures.append(run_bash_command_with_log_and_return_pid \\
                    .options(name=name_str, {options_str}) \\
                    .remote(
                        script,
                        log_path,
                        env_vars=sky_env_vars_dict,
                        stream_logs=True,
                        with_ray=True,
                    ))""")
        ]

    def add_epilogue(self) -> None:
        """Generates code that waits for all tasks, then exits."""
        self._code.append('returncodes, _ = get_or_fail(futures, pg)')
        super().add_epilogue()


class SlurmCodeGen(TaskCodeGen):
    """Code generator for task execution on Slurm using native srun."""

    def __init__(
        self,
        slurm_job_id: str,
        container_name: Optional[str],
    ):
        """Initialize SlurmCodeGen.

        Args:
            slurm_job_id: The Slurm job ID, i.e. SLURM_JOB_ID
            container_name: pyxis container name, or None
        """
        super().__init__()
        self._slurm_job_id = slurm_job_id
        self._container_name = container_name

    def add_prologue(self, job_id: int) -> None:
        assert not self._has_prologue, 'add_prologue() called twice?'
        self._has_prologue = True
        self.job_id = job_id

        self._add_common_imports()

        self._code.append(
            textwrap.dedent("""\
            import colorama
            import copy
            import json
            import multiprocessing
            import signal
            import threading
            from sky.backends import backend_utils
            """))
        self._add_skylet_imports()

        self._add_constants()

        self._add_logging_functions()

        self._code.append(
            textwrap.dedent(f"""\
            def _cancel_slurm_job_steps():
                slurm_job_id = {self._slurm_job_id!r}
                assert slurm_job_id is not None, 'SLURM_JOB_ID is not set'
                try:
                    # Query steps for this job: squeue -s -j JOBID -h -o "%i %j"
                    # Output format: "JOBID.STEPID STEPNAME"
                    # TODO(kevin): This assumes that compute node is able
                    # to run client commands against the controller.
                    # Validate this assumption.
                    result = subprocess.run(
                        ['squeue', '-s', '-j', slurm_job_id, '-h', '-o', '%i %j'],
                        capture_output=True, text=True, check=False)
                    for line in result.stdout.strip().split('\\n'):
                        if not line:
                            continue
                        parts = line.split()
                        assert len(parts) >= 2, 'Expected at least 2 parts'
                        step_id, step_name = parts[0], parts[1]
                        if step_name == f'sky-{self.job_id}':
                            subprocess.run(['scancel', step_id],
                                            check=False, capture_output=True)
                except Exception as e:
                    print(f'Error in _cancel_slurm_job_steps: {{e}}', flush=True)
                    pass

            def _slurm_cleanup_handler(signum, _frame):
                _cancel_slurm_job_steps()
                # Re-raise to let default handler terminate.
                signal.signal(signum, signal.SIG_DFL)
                os.kill(os.getpid(), signum)

            signal.signal(signal.SIGTERM, _slurm_cleanup_handler)
            """))

        self._code += [
            'autostop_lib.set_last_active_time_to_now()',
            f'job_lib.set_status({job_id!r}, job_lib.JobStatus.PENDING)',
        ]

        self._setup_cmd: Optional[str] = None
        self._setup_envs: Optional[Dict[str, str]] = None
        self._setup_log_dir: Optional[str] = None
        self._setup_num_nodes: Optional[int] = None

    def add_setup(
        self,
        num_nodes: int,
        resources_dict: Dict[str, float],
        stable_cluster_internal_ips: List[str],
        env_vars: Dict[str, str],
        log_dir: str,
        setup_cmd: Optional[str] = None,
    ) -> None:
        assert self._has_prologue, ('Call add_prologue() before add_setup().')
        self._has_setup = True
        self._cluster_num_nodes = len(stable_cluster_internal_ips)
        self._stable_cluster_ips = stable_cluster_internal_ips

        self._add_waiting_for_resources_msg(num_nodes)

        # Store setup information for use in add_task().
        if setup_cmd is not None:
            setup_envs = env_vars.copy()
            setup_envs[constants.SKYPILOT_NUM_NODES] = str(num_nodes)
            self._setup_cmd = setup_cmd
            self._setup_envs = setup_envs
            self._setup_log_dir = log_dir
            self._setup_num_nodes = num_nodes

    def add_task(
        self,
        num_nodes: int,
        bash_script: Optional[str],
        task_name: Optional[str],
        resources_dict: Dict[str, float],
        log_dir: str,
        env_vars: Optional[Dict[str, str]] = None,
    ) -> None:
        """Generates code for invoking a bash command
        using srun within sbatch allocation.
        """
        assert self._has_setup, 'Call add_setup() before add_task().'
        env_vars = env_vars or {}
        task_name = task_name if task_name is not None else 'task'

        acc_name, acc_count = self._get_accelerator_details(resources_dict)
        num_gpus = 0
        if (acc_name is not None and
                not accelerator_registry.is_schedulable_non_gpu_accelerator(
                    acc_name)):
            num_gpus = int(math.ceil(acc_count))

        # Slurm does not support fractional CPUs.
        task_cpu_demand = int(math.ceil(resources_dict.pop('CPU')))

        sky_env_vars_dict_str = [
            textwrap.dedent(f"""\
            sky_env_vars_dict = {{}}
            sky_env_vars_dict['SKYPILOT_INTERNAL_JOB_ID'] = {self.job_id}
            """)
        ]

        if env_vars:
            sky_env_vars_dict_str.extend(f'sky_env_vars_dict[{k!r}] = {v!r}'
                                         for k, v in env_vars.items())
        sky_env_vars_dict_str = '\n'.join(sky_env_vars_dict_str)

        rclone_flush_script = self.get_rclone_flush_script()
        streaming_msg = self._get_job_started_msg()
        has_setup_cmd = self._setup_cmd is not None

        container_flags = ''
        if self._container_name is not None:
            # --container-remap-root must be passed on every srun to get
            # correct $HOME
            container_flags = (
                ' --container-remap-root'
                f' --container-name={shlex.quote(self._container_name)}:exec')

        self._code += [
            sky_env_vars_dict_str,
            textwrap.dedent(f"""\
            script = {bash_script!r}
            if script is None:
                script = ''
            rclone_flush_script = {rclone_flush_script!r}

            if script or {has_setup_cmd!r}:
                script += rclone_flush_script
                sky_env_vars_dict['{constants.SKYPILOT_NUM_GPUS_PER_NODE}'] = {num_gpus}

                # Signal files for setup/run synchronization:
                # 1. alloc_signal_file: srun has acquired allocation
                # 2. setup_done_signal_file: Driver has finished setup, run can proceed
                #
                # Signal files are stored in home directory, which is
                # assumed to be on a shared NFS mount accessible by all nodes.
                # To support clusters with non-NFS home directories, we would
                # need to let users specify an NFS-backed "working directory"
                # or use a different coordination mechanism.
                alloc_signal_file = f'~/.sky_alloc_{self._slurm_job_id}_{self.job_id}'
                alloc_signal_file = os.path.expanduser(alloc_signal_file)
                setup_done_signal_file = f'~/.sky_setup_done_{self._slurm_job_id}_{self.job_id}'
                setup_done_signal_file = os.path.expanduser(setup_done_signal_file)

                # Start exclusive srun in a thread to reserve allocation (similar to ray.get(pg.ready()))
                gpu_arg = f'--gpus-per-node={num_gpus}' if {num_gpus} > 0 else ''

                def build_task_runner_cmd(user_script, extra_flags, log_dir, env_vars_dict,
                                          task_name=None, is_setup=False,
                                          alloc_signal=None, setup_done_signal=None):
                    env_vars_json = json.dumps(env_vars_dict)

                    log_dir = shlex.quote(log_dir)
                    env_vars = shlex.quote(env_vars_json)
                    cluster_ips = shlex.quote(",".join({self._stable_cluster_ips!r}))

                    runner_args = f'--log-dir={{log_dir}} --env-vars={{env_vars}} --cluster-num-nodes={self._cluster_num_nodes} --cluster-ips={{cluster_ips}}'

                    if task_name is not None:
                        runner_args += f' --task-name={{shlex.quote(task_name)}}'

                    if is_setup:
                        runner_args += ' --is-setup'

                    if alloc_signal is not None:
                        runner_args += f' --alloc-signal-file={{shlex.quote(alloc_signal)}}'

                    if setup_done_signal is not None:
                        runner_args += f' --setup-done-signal-file={{shlex.quote(setup_done_signal)}}'

                    script_path = None
                    prefix = 'sky_setup_' if is_setup else 'sky_task_'
                    if backend_utils.is_command_length_over_limit(user_script):
                        with tempfile.NamedTemporaryFile('w', prefix=prefix, suffix='.sh', delete=False) as f:
                            f.write(user_script)
                            script_path = f.name
                        runner_args += f' --script-path={{shlex.quote(script_path)}}'
                    else:
                        runner_args += f' --script={{shlex.quote(user_script)}}'

                    # Use /usr/bin/env explicitly to work around a Slurm quirk where
                    # srun's execvp() doesn't check execute permissions, failing when
                    # $HOME/.local/bin/env (non-executable, from uv installation)
                    # shadows /usr/bin/env.
                    job_suffix = '-setup' if is_setup else ''
                    # Unset SLURM_* environment variables before running srun.
                    # When this srun runs inside another srun (from
                    # SlurmCommandRunner.run), inherited variables like
                    # SLURM_CPU_BIND, SLURM_NNODES, and SLURM_NODELIST constrain
                    # the inner srun to the parent step's allocation. This causes
                    # "CPU binding outside of job step allocation" errors.
                    # Unsetting all SLURM_* variables allows this srun to access the full job
                    # allocation. See:
                    # https://support.schedmd.com/show_bug.cgi?id=14298
                    # https://github.com/huggingface/datatrove/issues/248
                    cmd_parts = []
                    # Only unset SKY_RUNTIME_DIR for container runs. For non-container
                    # runs, we want to inherit the node-local SKY_RUNTIME_DIR set by
                    # SlurmCommandRunner to avoid SQLite WAL issues on shared filesystems.
                    if {True if container_flags else False}:
                        cmd_parts.append('unset SKY_RUNTIME_DIR;')
                    cmd_parts.extend([
                        constants.SKY_SLURM_PYTHON_CMD,
                        '-m sky.skylet.executor.slurm',
                        runner_args,
                    ])
                    bash_cmd = shlex.quote(' '.join(cmd_parts))
                    srun_cmd = (
                        "unset $(env | awk -F= '/^SLURM_/ {{print $1}}') && "
                        f'srun --export=ALL --quiet --unbuffered --kill-on-bad-exit --jobid={self._slurm_job_id} '
                        f'--job-name=sky-{self.job_id}{{job_suffix}} --ntasks-per-node=1{container_flags} {{extra_flags}} '
                        f'/bin/bash -c {{bash_cmd}}'
                    )

                    def cleanup():
                        if script_path is not None:
                            os.remove(script_path)

                    return srun_cmd, cleanup

                def run_thread_func():
                    # This blocks until Slurm allocates resources (--exclusive)
                    # --mem=0 to match RayCodeGen's behavior where we don't explicitly request memory.
                    run_flags = f'--nodes={num_nodes} --cpus-per-task={task_cpu_demand} --mem=0 {{gpu_arg}} --exclusive'
                    srun_cmd, cleanup = build_task_runner_cmd(
                        script, run_flags, {log_dir!r}, sky_env_vars_dict,
                        task_name={task_name!r},
                        alloc_signal=alloc_signal_file,
                        setup_done_signal=setup_done_signal_file
                    )

                    proc = subprocess.Popen(srun_cmd, shell=True,
                                          stdout=subprocess.PIPE,
                                          stderr=subprocess.STDOUT,
                                          text=True)
                    for line in proc.stdout:
                        print(line, end='', flush=True)
                    proc.wait()

                    cleanup()
                    return {{'return_code': proc.returncode, 'pid': proc.pid}}

                run_thread_result = {{'result': None}}
                def run_thread_wrapper():
                    run_thread_result['result'] = run_thread_func()

                run_thread = threading.Thread(target=run_thread_wrapper)
                run_thread.start()

                # Wait for allocation signal from inside srun
                while not os.path.exists(alloc_signal_file):
                    if not run_thread.is_alive():
                        # srun failed before creating the signal file.
                        run_thread.join()
                        result = run_thread_result['result']
                        returncode = int(result.get('return_code', 1))
                        pid = result.get('pid', os.getpid())
                        msg = f'ERROR: {colorama.Fore.RED}Job {self.job_id}\\'s setup failed with return code {{returncode}} (pid={{pid}}).'
                        msg += f' See error logs above for more details.{colorama.Style.RESET_ALL}'
                        print(msg, flush=True)
                        returncodes = [returncode]
                        if int(constants.SKYLET_VERSION) >= 28:
                            job_lib.set_exit_codes({self.job_id!r}, returncodes)
                        job_lib.set_status({self.job_id!r}, job_lib.JobStatus.FAILED_SETUP)
                        sys.exit(1)
                    time.sleep(0.1)

                print({streaming_msg!r}, flush=True)

                if {has_setup_cmd!r}:
                    job_lib.set_status({self.job_id!r}, job_lib.JobStatus.SETTING_UP)

                    # The schedule_step should be called after the job status is set to
                    # non-PENDING, otherwise, the scheduler will think the current job
                    # is not submitted yet, and skip the scheduling step.
                    job_lib.scheduler.schedule_step()

                    # --overlap as we have already secured allocation with the srun for the run section,
                    # and otherwise this srun would get blocked and deadlock.
                    setup_flags = f'--overlap --nodes={self._setup_num_nodes}'
                    setup_srun, setup_cleanup = build_task_runner_cmd(
                        {self._setup_cmd!r}, setup_flags, {self._setup_log_dir!r}, {self._setup_envs!r},
                        is_setup=True
                    )

                    # Run setup srun directly, streaming output to driver stdout
                    setup_proc = subprocess.Popen(setup_srun, shell=True,
                                                 stdout=subprocess.PIPE,
                                                 stderr=subprocess.STDOUT,
                                                 text=True)
                    for line in setup_proc.stdout:
                        print(line, end='', flush=True)
                    setup_proc.wait()

                    setup_cleanup()

                    setup_returncode = setup_proc.returncode
                    if setup_returncode != 0:
                        setup_pid = setup_proc.pid
                        msg = f'ERROR: {colorama.Fore.RED}Job {self.job_id}\\'s setup failed with return code {{setup_returncode}} (pid={{setup_pid}}).'
                        msg += f' See error logs above for more details.{colorama.Style.RESET_ALL}'
                        print(msg, flush=True)
                        job_lib.set_status({self.job_id!r}, job_lib.JobStatus.FAILED_SETUP)
                        # Cancel the srun spawned by run_thread_func.
                        _cancel_slurm_job_steps()
                        sys.exit(1)

                job_lib.set_job_started({self.job_id!r})
                if not {has_setup_cmd!r}:
                    # Need to call schedule_step() to make sure the scheduler
                    # schedule the next pending job.
                    job_lib.scheduler.schedule_step()

                # Signal run thread to proceed.
                pathlib.Path(setup_done_signal_file).touch()

                # Wait for run thread to complete.
                run_thread.join()
                result = run_thread_result['result']

                # Cleanup signal files
                if os.path.exists(alloc_signal_file):
                    os.remove(alloc_signal_file)
                if os.path.exists(setup_done_signal_file):
                    os.remove(setup_done_signal_file)

                returncodes = [int(result.get('return_code', 1))]
            else:
                returncodes = [0]
            """),
        ]
